# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base grid classes
"""

import functools
import json
import logging
import warnings
from collections import namedtuple, OrderedDict

import sqlalchemy as sa
from sqlalchemy import orm

import paginate
from paginate_sqlalchemy import SqlalchemyOrmPage
from pyramid.renderers import render
from webhelpers2.html import HTML

from wuttaweb.db import Session
from wuttaweb.util import FieldList, get_model_fields, make_json_safe
from wuttjamaican.util import UNSPECIFIED
from wuttaweb.grids.filters import default_sqlalchemy_filters, VerbNotSupported


log = logging.getLogger(__name__)


SortInfo = namedtuple('SortInfo', ['sortkey', 'sortdir'])
SortInfo.__doc__ = """
Named tuple to track sorting info.

Elements of :attr:`~Grid.sort_defaults` will be of this type.
"""

class Grid:
    """
    Base class for all :term:`grids <grid>`.

    :param request: Reference to current :term:`request` object.

    :param columns: List of column names for the grid.  This is
       optional; if not specified an attempt will be made to deduce
       the list automatically.  See also :attr:`columns`.

    .. note::

       Some parameters are not explicitly described above.  However
       their corresponding attributes are described below.

    Grid instances contain the following attributes:

    .. attribute:: key

       Presumably unique key for the grid; used to track per-grid
       sort/filter settings etc.

    .. attribute:: vue_tagname

       String name for Vue component tag.  By default this is
       ``'wutta-grid'``.  See also :meth:`render_vue_tag()`.

    .. attribute:: model_class

       Model class for the grid, if applicable.  When set, this is
       usually a SQLAlchemy mapped class.  This may be used for
       deriving the default :attr:`columns` among other things.

    .. attribute:: columns

       :class:`~wuttaweb.util.FieldList` instance containing string
       column names for the grid.  Columns will appear in the same
       order as they are in this list.

       See also :meth:`set_columns()` and :meth:`get_columns()`.

    .. attribute:: data

       Data set for the grid.  This should either be a list of dicts
       (or objects with dict-like access to fields, corresponding to
       model records) or else an object capable of producing such a
       list, e.g. SQLAlchemy query.

       This is the "full" data set; see also
       :meth:`get_visible_data()`.

    .. attribute:: labels

       Dict of column label overrides.

       See also :meth:`get_label()` and :meth:`set_label()`.

    .. attribute:: renderers

       Dict of column (cell) value renderer overrides.

       See also :meth:`set_renderer()` and
       :meth:`set_default_renderers()`.

    .. attribute:: row_class

       This represents the CSS ``class`` attribute for a row within
       the grid.  Default is ``None``.

       This can be a simple string, in which case the same class is
       applied to all rows.

       Or it can be a callable, which can then return different
       class(es) depending on each row.  The callable must take three
       args: ``(obj, data, i)`` - for example::

          def my_row_class(obj, data, i):
              if obj.archived:
                  return 'poser-archived'

          grid = Grid(request, key='foo', row_class=my_row_class)

       See :meth:`get_row_class()` for more info.

    .. attribute:: actions

       List of :class:`GridAction` instances represenging action links
       to be shown for each record in the grid.

    .. attribute:: linked_columns

       List of column names for which auto-link behavior should be
       applied.

       See also :meth:`set_link()` and :meth:`is_linked()`.

    .. attribute:: sortable

       Boolean indicating whether *any* column sorting is allowed for
       the grid.  Default is ``False``.

       See also :attr:`sort_multiple` and :attr:`sort_on_backend`.

    .. attribute:: sort_multiple

       Boolean indicating whether "multi-column" sorting is allowed.
       Default is ``True``; if this is ``False`` then only one column
       may be sorted at a time.

       Only relevant if :attr:`sortable` is true, but applies to both
       frontend and backend sorting.

       .. warning::

          This feature is limited by frontend JS capabilities,
          regardless of :attr:`sort_on_backend` value (i.e. for both
          frontend and backend sorting).

          In particular, if the app theme templates use Vue 2 + Buefy,
          then multi-column sorting should work.

          But not so with Vue 3 + Oruga, *yet* - see also the `open
          issue <https://github.com/oruga-ui/oruga/issues/962>`_
          regarding that.  For now this flag is simply ignored for
          Vue 3 + Oruga templates.

          Additionally, even with Vue 2 + Buefy this flag can only
          allow the user to *request* a multi-column sort.  Whereas
          the "default sort" in the Vue component can only ever be
          single-column, regardless of :attr:`sort_defaults`.

    .. attribute:: sort_on_backend

       Boolean indicating whether the grid data should be sorted on the
       backend.  Default is ``True``.

       If ``False``, the client-side Vue component will handle the
       sorting.

       Only relevant if :attr:`sortable` is also true.

    .. attribute:: sorters

       Dict of functions to use for backend sorting.

       Only relevant if both :attr:`sortable` and
       :attr:`sort_on_backend` are true.

       See also :meth:`set_sorter()`, :attr:`sort_defaults` and
       :attr:`active_sorters`.

    .. attribute:: sort_defaults

       List of options to be used for default sorting, until the user
       requests a different sorting method.

       This list usually contains either zero or one elements.  (More
       are allowed if :attr:`sort_multiple` is true, but see note
       below.)  Each list element is a :class:`SortInfo` tuple and
       must correspond to an entry in :attr:`sorters`.

       Used with both frontend and backend sorting.

       See also :meth:`set_sort_defaults()` and
       :attr:`active_sorters`.

       .. warning::

          While the grid logic is built to handle multi-column
          sorting, this feature is limited by frontend JS
          capabilities.

          Even if ``sort_defaults`` contains multiple entries
          (i.e. for multi-column sorting to be used "by default" for
          the grid), only the *first* entry (i.e. single-column
          sorting) will actually be used as the default for the Vue
          component.

          See also :attr:`sort_multiple` for more details.

    .. attribute:: active_sorters

       List of sorters currently in effect for the grid; used by
       :meth:`sort_data()`.

       Whereas :attr:`sorters` defines all "available" sorters, and
       :attr:`sort_defaults` defines the "default" sorters,
       ``active_sorters`` defines the "current/effective" sorters.

       This attribute is set by :meth:`load_settings()`; until that is
       called it will not exist.

       This is conceptually a "subset" of :attr:`sorters` although a
       different format is used here::

          grid.active_sorters = [
              {'key': 'name', 'dir': 'asc'},
              {'key': 'id', 'dir': 'asc'},
          ]

       The above is for example only; there is usually no reason to
       set this attribute directly.

       This list may contain multiple elements only if
       :attr:`sort_multiple` is true.  Otherewise it should always
       have either zero or one element.

    .. attribute:: paginated

       Boolean indicating whether the grid data should be paginated,
       i.e. split up into pages.  Default is ``False`` which means all
       data is shown at once.

       See also :attr:`pagesize` and :attr:`page`, and
       :attr:`paginate_on_backend`.

    .. attribute:: paginate_on_backend

       Boolean indicating whether the grid data should be paginated on
       the backend.  Default is ``True`` which means only one "page"
       of data is sent to the client-side component.

       If this is ``False``, the full set of grid data is sent for
       each request, and the client-side Vue component will handle the
       pagination.

       Only relevant if :attr:`paginated` is also true.

    .. attribute:: pagesize_options

       List of "page size" options for the grid.  See also
       :attr:`pagesize`.

       Only relevant if :attr:`paginated` is true.  If not specified,
       constructor will call :meth:`get_pagesize_options()` to get the
       value.

    .. attribute:: pagesize

       Number of records to show in a data page.  See also
       :attr:`pagesize_options` and :attr:`page`.

       Only relevant if :attr:`paginated` is true.  If not specified,
       constructor will call :meth:`get_pagesize()` to get the value.

    .. attribute:: page

       The current page number (of data) to display in the grid.  See
       also :attr:`pagesize`.

       Only relevant if :attr:`paginated` is true.  If not specified,
       constructor will assume ``1`` (first page).

    .. attribute:: searchable_columns

       Set of columns declared as searchable for the Vue component.

       See also :meth:`set_searchable()` and :meth:`is_searchable()`.

    .. attribute:: filterable

       Boolean indicating whether the grid should show a "filters"
       section where user can filter data in various ways.  Default is
       ``False``.

    .. attribute:: filters

       Dict of :class:`~wuttaweb.grids.filters.GridFilter` instances
       available for use with backend filtering.

       Only relevant if :attr:`filterable` is true.

       See also :meth:`set_filter()`.

    .. attribute:: filter_defaults

       Dict containing default state preferences for the filters.

       See also :meth:`set_filter_defaults()`.

    .. attribute:: joiners

       Dict of "joiner" functions for use with backend filtering and
       sorting.

       See :meth:`set_joiner()` for more info.

    .. attribute:: tools

       Dict of "tool" elements for the grid.  Tools are usually buttons
       (e.g. "Delete Results"), shown on top right of the grid.

       The keys for this dict are somewhat arbitrary, defined by the
       caller.  Values should be HTML literal elements.

       See also :meth:`add_tool()` and :meth:`set_tools()`.
    """

    def __init__(
            self,
            request,
            vue_tagname='wutta-grid',
            model_class=None,
            key=None,
            columns=None,
            data=None,
            labels={},
            renderers={},
            row_class=None,
            actions=[],
            linked_columns=[],
            sortable=False,
            sort_multiple=True,
            sort_on_backend=True,
            sorters=None,
            sort_defaults=None,
            paginated=False,
            paginate_on_backend=True,
            pagesize_options=None,
            pagesize=None,
            page=1,
            searchable_columns=None,
            filterable=False,
            filters=None,
            filter_defaults=None,
            joiners=None,
            tools=None,
    ):
        self.request = request
        self.vue_tagname = vue_tagname
        self.model_class = model_class
        self.key = key
        self.data = data
        self.labels = labels or {}
        self.row_class = row_class
        self.actions = actions or []
        self.linked_columns = linked_columns or []
        self.joiners = joiners or {}

        self.config = self.request.wutta_config
        self.app = self.config.get_app()

        self.set_columns(columns or self.get_columns())
        self.renderers = {}
        if renderers:
            for key, val in renderers.items():
                self.set_renderer(key, val)
        self.set_default_renderers()
        self.set_tools(tools)

        # sorting
        self.sortable = sortable
        self.sort_multiple = sort_multiple
        if self.sort_multiple and self.request.use_oruga:
            log.warning("grid.sort_multiple is not implemented for Oruga-based templates")
            self.sort_multiple = False
        self.sort_on_backend = sort_on_backend
        if sorters is not None:
            self.sorters = sorters
        elif self.sortable and self.sort_on_backend:
            self.sorters = self.make_backend_sorters()
        else:
            self.sorters = {}
        self.set_sort_defaults(sort_defaults or [])

        # paging
        self.paginated = paginated
        self.paginate_on_backend = paginate_on_backend
        self.pagesize_options = pagesize_options or self.get_pagesize_options()
        self.pagesize = pagesize or self.get_pagesize()
        self.page = page

        # searching
        self.searchable_columns = set(searchable_columns or [])

        # filtering
        self.filterable = filterable
        if filters is not None:
            self.filters = filters
        elif self.filterable:
            self.filters = self.make_backend_filters()
        else:
            self.filters = {}
        self.set_filter_defaults(**(filter_defaults or {}))

    def get_columns(self):
        """
        Returns the official list of column names for the grid, or
        ``None``.

        If :attr:`columns` is set and non-empty, it is returned.

        Or, if :attr:`model_class` is set, the field list is derived
        from that, via :meth:`get_model_columns()`.

        Otherwise ``None`` is returned.
        """
        if hasattr(self, 'columns') and self.columns:
            return self.columns

        columns = self.get_model_columns()
        if columns:
            return columns

        return []

    def get_model_columns(self, model_class=None):
        """
        This method is a shortcut which calls
        :func:`~wuttaweb.util.get_model_fields()`.

        :param model_class: Optional model class for which to return
           fields.  If not set, the grid's :attr:`model_class` is
           assumed.
        """
        return get_model_fields(self.config,
                                model_class=model_class or self.model_class)

    @property
    def vue_component(self):
        """
        String name for the Vue component, e.g. ``'WuttaGrid'``.

        This is a generated value based on :attr:`vue_tagname`.
        """
        words = self.vue_tagname.split('-')
        return ''.join([word.capitalize() for word in words])

    def set_columns(self, columns):
        """
        Explicitly set the list of grid columns.

        This will overwrite :attr:`columns` with a new
        :class:`~wuttaweb.util.FieldList` instance.

        :param columns: List of string column names.
        """
        self.columns = FieldList(columns)

    def append(self, *keys):
        """
        Add some columns(s) to the grid.

        This is a convenience to allow adding multiple columns at
        once::

           grid.append('first_field',
                       'second_field',
                       'third_field')

        It will add each column to :attr:`columns`.
        """
        for key in keys:
            if key not in self.columns:
                self.columns.append(key)

    def remove(self, *keys):
        """
        Remove some column(s) from the grid.

        This is a convenience to allow removal of multiple columns at
        once::

           grid.remove('first_field',
                       'second_field',
                       'third_field')

        It will remove each column from :attr:`columns`.
        """
        for key in keys:
            if key in self.columns:
                self.columns.remove(key)

    def set_label(self, key, label, column_only=False):
        """
        Set/override the label for a column.

        :param key: Name of column.

        :param label: New label for the column header.

        :param column_only: Boolean indicating whether the label
           should be applied *only* to the column header (if
           ``True``), vs.  applying also to the filter (if ``False``).

        See also :meth:`get_label()`.  Label overrides are tracked via
        :attr:`labels`.
        """
        self.labels[key] = label

        if not column_only and key in self.filters:
            self.filters[key].label = label

    def get_label(self, key):
        """
        Returns the label text for a given column.

        If no override is defined, the label is derived from ``key``.

        See also :meth:`set_label()`.
        """
        if key in self.labels:
            return self.labels[key]
        return self.app.make_title(key)

    def set_renderer(self, key, renderer, **kwargs):
        """
        Set/override the value renderer for a column.

        :param key: Name of column.

        :param renderer: Callable as described below.

        Depending on the nature of grid data, sometimes a cell's
        "as-is" value will be undesirable for display purposes.

        The logic in :meth:`get_vue_context()` will first "convert"
        all grid data as necessary so that it is at least
        JSON-compatible.

        But then it also will invoke a renderer override (if defined)
        to obtain the "final" cell value.

        A renderer must be a callable which accepts 3 args ``(record,
        key, value)``:

        * ``record`` is the "original" record from :attr:`data`
        * ``key`` is the column name
        * ``value`` is the JSON-safe cell value

        Whatever the renderer returns, is then used as final cell
        value.  For instance::

           from webhelpers2.html import HTML

           def render_foo(record, key, value):
              return HTML.literal("<p>this is the final cell value</p>")

           grid = Grid(request, columns=['foo', 'bar'])
           grid.set_renderer('foo', render_foo)

        For convenience, in lieu of a renderer callable, you may
        specify one of the following strings, which will be
        interpreted as a built-in renderer callable, as shown below:

        * ``'batch_id'`` -> :meth:`render_batch_id()`
        * ``'boolean'`` -> :meth:`render_boolean()`
        * ``'currency'`` -> :meth:`render_currency()`
        * ``'date'`` -> :meth:`render_date()`
        * ``'datetime'`` -> :meth:`render_datetime()`
        * ``'quantity'`` -> :meth:`render_quantity()`

        Renderer overrides are tracked via :attr:`renderers`.
        """
        builtins = {
            'batch_id': self.render_batch_id,
            'boolean': self.render_boolean,
            'currency': self.render_currency,
            'date': self.render_date,
            'datetime': self.render_datetime,
            'quantity': self.render_quantity,
        }

        if renderer in builtins:
            renderer = builtins[renderer]

        if kwargs:
            renderer = functools.partial(renderer, **kwargs)
        self.renderers[key] = renderer

    def set_default_renderers(self):
        """
        Set default column value renderers, where applicable.

        This is called automatically from the class constructor.  It
        will add new entries to :attr:`renderers` for columns whose
        data type implies a default renderer.  This is only possible
        if :attr:`model_class` is set to a SQLAlchemy mapped class.

        This only looks for a few data types, and configures as
        follows:

        * :class:`sqlalchemy:sqlalchemy.types.Boolean` ->
          :meth:`render_boolean()`
        * :class:`sqlalchemy:sqlalchemy.types.Date` ->
          :meth:`render_date()`
        * :class:`sqlalchemy:sqlalchemy.types.DateTime` ->
          :meth:`render_datetime()`
        """
        if not self.model_class:
            return

        for key in self.columns:
            if key in self.renderers:
                continue

            attr = getattr(self.model_class, key, None)
            if attr:
                prop = getattr(attr, 'prop', None)
                if prop and isinstance(prop, orm.ColumnProperty):
                    column = prop.columns[0]
                    if isinstance(column.type, sa.Date):
                        self.set_renderer(key, self.render_date)
                    elif isinstance(column.type, sa.DateTime):
                        self.set_renderer(key, self.render_datetime)
                    elif isinstance(column.type, sa.Boolean):
                        self.set_renderer(key, self.render_boolean)

    def set_link(self, key, link=True):
        """
        Explicitly enable or disable auto-link behavior for a given
        column.

        If a column has auto-link enabled, then each of its cell
        contents will automatically be wrapped with a hyperlink.  The
        URL for this will be the same as for the "View"
        :class:`GridAction`
        (aka. :meth:`~wuttaweb.views.master.MasterView.view()`).
        Although of course each cell in the column gets a different
        link depending on which data record it points to.

        It is typical to enable auto-link for fields relating to ID,
        description etc. or some may prefer to auto-link all columns.

        See also :meth:`is_linked()`; the list is tracked via
        :attr:`linked_columns`.

        :param key: Column key as string.

        :param link: Boolean indicating whether column's cell contents
           should be auto-linked.
        """
        if link:
            if key not in self.linked_columns:
                self.linked_columns.append(key)
        else: # unlink
            if self.linked_columns and key in self.linked_columns:
                self.linked_columns.remove(key)

    def is_linked(self, key):
        """
        Returns boolean indicating if auto-link behavior is enabled
        for a given column.

        See also :meth:`set_link()` which describes auto-link behavior.

        :param key: Column key as string.
        """
        if self.linked_columns:
            if key in self.linked_columns:
                return True
        return False

    def set_searchable(self, key, searchable=True):
        """
        (Un)set the given column's searchable flag for the Vue
        component.

        See also :meth:`is_searchable()`.  Flags are tracked via
        :attr:`searchable_columns`.
        """
        if searchable:
            self.searchable_columns.add(key)
        elif key in self.searchable_columns:
            self.searchable_columns.remove(key)

    def is_searchable(self, key):
        """
        Check if the given column is marked as searchable for the Vue
        component.

        See also :meth:`set_searchable()`.
        """
        return key in self.searchable_columns

    def add_action(self, key, **kwargs):
        """
        Convenience to add a new :class:`GridAction` instance to the
        grid's :attr:`actions` list.
        """
        self.actions.append(GridAction(self.request, key, **kwargs))

    def set_tools(self, tools):
        """
        Set the :attr:`tools` attribute using the given tools collection.

        This will normalize the list/dict to desired internal format.
        """
        if tools and isinstance(tools, list):
            if not any([isinstance(t, (tuple, list)) for t in tools]):
                tools = [(self.app.make_uuid(), t) for t in tools]
        self.tools = OrderedDict(tools or [])

    def add_tool(self, html, key=None):
        """
        Add a new HTML snippet to the :attr:`tools` dict.

        :param html: HTML literal for the tool element.

        :param key: Optional key to use when adding to the
           :attr:`tools` dict.  If not specified, a random string is
           generated.

        See also :meth:`set_tools()`.
        """
        if not key:
            key = self.app.make_uuid()
        self.tools[key] = html

    ##############################
    # joining methods
    ##############################

    def set_joiner(self, key, joiner):
        """
        Set/override the backend joiner for a column.

        A "joiner" is sometimes needed when a column with "related but
        not primary" data is involved in a sort or filter operation.

        A sorter or filter may need to "join" other table(s) to get at
        the appropriate data.  But if a given column has both a sorter
        and filter defined, and both are used at the same time, we
        don't want the join to happen twice.

        Hence we track joiners separately, also keyed by column name
        (as are sorters and filters).  When a column's sorter **and/or**
        filter is needed, the joiner will be invoked.

        :param key: Name of column.

        :param joiner: A joiner callable, as described below.

        A joiner callable must accept just one ``(data)`` arg and
        return the "joined" data/query, for example::

           model = app.model
           grid = Grid(request, model_class=model.Person)

           def join_external_profile_value(query):
               return query.join(model.ExternalProfile)

           def sort_external_profile(query, direction):
              sortspec = getattr(model.ExternalProfile.description, direction)
              return query.order_by(sortspec())

           grid.set_joiner('external_profile', join_external_profile)
           grid.set_sorter('external_profile', sort_external_profile)

        See also :meth:`remove_joiner()`.  Backend joiners are tracked
        via :attr:`joiners`.
        """
        self.joiners[key] = joiner

    def remove_joiner(self, key):
        """
        Remove the backend joiner for a column.

        Note that this removes the joiner *function*, so there is no
        way to apply joins for this column unless another joiner is
        later defined for it.

        See also :meth:`set_joiner()`.
        """
        self.joiners.pop(key, None)

    ##############################
    # sorting methods
    ##############################

    def make_backend_sorters(self, sorters=None):
        """
        Make backend sorters for all columns in the grid.

        This is called by the constructor, if both :attr:`sortable`
        and :attr:`sort_on_backend` are true.

        For each column in the grid, this checks the provided
        ``sorters`` and if the column is not yet in there, will call
        :meth:`make_sorter()` to add it.

        .. note::

           This only works if grid has a :attr:`model_class`.  If not,
           this method just returns the initial sorters (or empty
           dict).

        :param sorters: Optional dict of initial sorters.  Any
           existing sorters will be left intact, not replaced.

        :returns: Final dict of all sorters.  Includes any from the
           initial ``sorters`` param as well as any which were
           created.
        """
        sorters = sorters or {}

        if self.model_class:
            for key in self.columns:
                if key in sorters:
                    continue
                prop = getattr(self.model_class, key, None)
                if (prop and hasattr(prop, 'property')
                    and isinstance(prop.property, orm.ColumnProperty)):
                    sorters[prop.key] = self.make_sorter(prop)

        return sorters

    def make_sorter(self, columninfo, keyfunc=None, foldcase=True):
        """
        Returns a function suitable for use as a backend sorter on the
        given column.

        Code usually does not need to call this directly.  See also
        :meth:`set_sorter()`, which calls this method automatically.

        :param columninfo: Can be either a model property (see below),
           or a column name.

        :param keyfunc: Optional function to use as the "sort key
           getter" callable, if the sorter is manual (as opposed to
           SQLAlchemy query).  More on this below.  If not specified,
           a default function is used.

        :param foldcase: If the sorter is manual (not SQLAlchemy), and
           the column data is of text type, this may be used to
           automatically "fold case" for the sorting.  Defaults to
           ``True`` since this behavior is presumably expected, but
           may be disabled if needed.

        The term "model property" is a bit technical, an example
        should help to clarify::

           model = app.model
           grid = Grid(request, model_class=model.Person)

           # explicit property
           sorter = grid.make_sorter(model.Person.full_name)

           # property name works if grid has model class
           sorter = grid.make_sorter('full_name')

           # nb. this will *not* work
           person = model.Person(full_name="John Doe")
           sorter = grid.make_sorter(person.full_name)

        The ``keyfunc`` param allows you to override the way sort keys
        are obtained from data records (this only applies for a
        "manual" sort, where data is a list and not a SQLAlchemy
        query)::

           data = [
                {'foo': 1},
                {'bar': 2},
           ]

           # nb. no model_class, just as an example
           grid = Grid(request, columns=['foo', 'bar'], data=data)

           def getkey(obj):
               if obj.get('foo')
                   return obj['foo']
               if obj.get('bar'):
                   return obj['bar']
               return ''

           # nb. sortfunc will ostensibly sort by 'foo' column, but in
           # practice it is sorted per value from getkey() above
           sortfunc = grid.make_sorter('foo', keyfunc=getkey)
           sorted_data = sortfunc(data, 'asc')

        :returns: A function suitable for backend sorting.  This
           function will behave differently when it is given a
           SQLAlchemy query vs. a "list" of data.  In either case it
           will return the sorted result.

           This function may be called as shown above.  It expects 2
           args: ``(data, direction)``
        """
        model_class = None
        model_property = None
        if isinstance(columninfo, str):
            key = columninfo
            model_class = self.model_class
            model_property = getattr(self.model_class, key, None)
        else:
            model_property = columninfo
            model_class = model_property.class_
            key = model_property.key

        def sorter(data, direction):

            # query is sorted with order_by()
            if isinstance(data, orm.Query):
                if not model_property:
                    raise TypeError(f"grid sorter for '{key}' does not map to a model property")
                query = data
                return query.order_by(getattr(model_property, direction)())

            # other data is sorted manually.  first step is to
            # identify the function used to produce a sort key for
            # each record
            kfunc = keyfunc
            if not kfunc:
                if model_property:
                    # TODO: may need this for String etc. as well?
                    if isinstance(model_property.type, sa.Text):
                        if foldcase:
                            kfunc = lambda obj: (obj[key] or '').lower()
                        else:
                            kfunc = lambda obj: obj[key] or ''
                if not kfunc:
                    # nb. sorting with this can raise error if data
                    # contains varying types, e.g. str and None
                    kfunc = lambda obj: obj[key]

            # then sort the data and return
            return sorted(data, key=kfunc, reverse=direction == 'desc')

        # TODO: this should be improved; is needed in tailbone for
        # multi-column sorting with sqlalchemy queries
        if model_property:
            sorter._class = model_class
            sorter._column = model_property

        return sorter

    def set_sorter(self, key, sortinfo=None):
        """
        Set/override the backend sorter for a column.

        Only relevant if both :attr:`sortable` and
        :attr:`sort_on_backend` are true.

        :param key: Name of column.

        :param sortinfo: Can be either a sorter callable, or else a
           model property (see below).

        If ``sortinfo`` is a callable, it will be used as-is for the
        backend sorter.

        Otherwise :meth:`make_sorter()` will be called to obtain the
        backend sorter.  The ``sortinfo`` will be passed along to that
        call; if it is empty then ``key`` will be used instead.

        A backend sorter callable must accept ``(data, direction)``
        args and return the sorted data/query, for example::

           model = app.model
           grid = Grid(request, model_class=model.Person)

           def sort_full_name(query, direction):
              sortspec = getattr(model.Person.full_name, direction)
              return query.order_by(sortspec())

           grid.set_sorter('full_name', sort_full_name)

        See also :meth:`remove_sorter()` and :meth:`is_sortable()`.
        Backend sorters are tracked via :attr:`sorters`.
        """
        sorter = None

        if sortinfo and callable(sortinfo):
            sorter = sortinfo
        else:
            sorter = self.make_sorter(sortinfo or key)

        self.sorters[key] = sorter

    def remove_sorter(self, key):
        """
        Remove the backend sorter for a column.

        Note that this removes the sorter *function*, so there is
        no way to sort by this column unless another sorter is
        later defined for it.

        See also :meth:`set_sorter()`.
        """
        self.sorters.pop(key, None)

    def set_sort_defaults(self, *args):
        """
        Set the default sorting method for the grid.  This sorting is
        used unless/until the user requests a different sorting
        method.

        ``args`` for this method are interpreted as follows:

        If 2 args are received, they should be for ``sortkey`` and
        ``sortdir``; for instance::

           grid.set_sort_defaults('name', 'asc')

        If just one 2-tuple arg is received, it is handled similarly::

           grid.set_sort_defaults(('name', 'asc'))

        If just one string arg is received, the default ``sortdir`` is
        assumed::

           grid.set_sort_defaults('name') # assumes 'asc'

        Otherwise there should be just one list arg, elements of
        which are each 2-tuples of ``(sortkey, sortdir)`` info::

           grid.set_sort_defaults([('name', 'asc'),
                                   ('value', 'desc')])

        .. note::

           Note that :attr:`sort_multiple` determines whether the grid
           is actually allowed to have multiple sort defaults.  The
           defaults requested by the method call may be pruned if
           necessary to accommodate that.

        Default sorting info is tracked via :attr:`sort_defaults`.
        """

        # convert args to sort defaults
        sort_defaults = []
        if len(args) == 1:
            if isinstance(args[0], str):
                sort_defaults = [SortInfo(args[0], 'asc')]
            elif isinstance(args[0], tuple) and len(args[0]) == 2:
                sort_defaults = [SortInfo(*args[0])]
            elif isinstance(args[0], list):
                sort_defaults = [SortInfo(*tup) for tup in args[0]]
            else:
                raise ValueError("for just one positional arg, must pass string, 2-tuple or list")
        elif len(args) == 2:
            sort_defaults = [SortInfo(*args)]
        else:
            raise ValueError("must pass just one or two positional args")

        # prune if multi-column requested but not supported
        if len(sort_defaults) > 1 and not self.sort_multiple:
            log.warning("multi-column sorting is not enabled for the instance; "
                        "list will be pruned to first element for '%s' grid: %s",
                        self.key, sort_defaults)
            sort_defaults = [sort_defaults[0]]

        self.sort_defaults = sort_defaults

    def is_sortable(self, key):
        """
        Returns boolean indicating if a given column should allow
        sorting.

        If :attr:`sortable` is false, this always returns ``False``.

        For frontend sorting (i.e. :attr:`sort_on_backend` is false),
        this always returns ``True``.

        For backend sorting, may return true or false depending on
        whether the column is listed in :attr:`sorters`.

        :param key: Column key as string.

        See also :meth:`set_sorter()`.
        """
        if not self.sortable:
            return False
        if self.sort_on_backend:
            return key in self.sorters
        return True

    ##############################
    # filtering methods
    ##############################

    def make_backend_filters(self, filters=None):
        """
        Make backend filters for all columns in the grid.

        This is called by the constructor, if :attr:`filterable` is
        true.

        For each column in the grid, this checks the provided
        ``filters`` and if the column is not yet in there, will call
        :meth:`make_filter()` to add it.

        .. note::

           This only works if grid has a :attr:`model_class`.  If not,
           this method just returns the initial filters (or empty
           dict).

        :param filters: Optional dict of initial filters.  Any
           existing filters will be left intact, not replaced.

        :returns: Final dict of all filters.  Includes any from the
           initial ``filters`` param as well as any which were
           created.
        """
        filters = filters or {}

        if self.model_class:

            # nb. i have found this confusing for some reason.  some
            # things i've tried so far include:
            #
            # i first tried self.get_model_columns() but my notes say
            # that was too aggressive in many cases.
            #
            # then i tried using the *subset* of self.columns, just
            # the ones which correspond to a property on the model
            # class.  but sometimes that skips filters we need.
            #
            # then i tried get_columns() from sa-utils to give the
            # "true" column list, but that fails when the underlying
            # column has different name than the prop/attr key.
            #
            # so now, we are looking directly at the sa mapper, for
            # all column attrs and then using the prop key.

            inspector = sa.inspect(self.model_class)
            for prop in inspector.column_attrs:
                if prop.key not in filters:
                    attr = getattr(self.model_class, prop.key)
                    filters[prop.key] = self.make_filter(attr)

        return filters

    def make_filter(self, columninfo, **kwargs):
        """
        Create and return a
        :class:`~wuttaweb.grids.filters.GridFilter` instance suitable
        for use on the given column.

        Code usually does not need to call this directly.  See also
        :meth:`set_filter()`, which calls this method automatically.

        :param columninfo: Can be either a model property (see below),
           or a column name.

        :returns: A :class:`~wuttaweb.grids.filters.GridFilter`
           instance.
        """
        key = kwargs.pop('key', None)

        # model_property is required
        model_property = None
        if kwargs.get('model_property'):
            model_property = kwargs['model_property']
        elif isinstance(columninfo, str):
            key = columninfo
            if self.model_class:
                model_property = getattr(self.model_class, key, None)
            if not model_property:
                raise ValueError(f"cannot locate model property for key: {key}")
        else:
            model_property = columninfo

        # optional factory override
        factory = kwargs.pop('factory', None)
        if not factory:
            typ = model_property.type
            factory = default_sqlalchemy_filters.get(type(typ))
            if not factory:
                factory = default_sqlalchemy_filters[None]

        # make filter
        kwargs['model_property'] = model_property
        return factory(self.request, key or model_property.key, **kwargs)

    def set_filter(self, key, filterinfo=None, **kwargs):
        """
        Set/override the backend filter for a column.

        Only relevant if :attr:`filterable` is true.

        :param key: Name of column.

        :param filterinfo: Can be either a
           :class:`~wuttweb.grids.filters.GridFilter` instance, or
           else a model property (see below).

        If ``filterinfo`` is a ``GridFilter`` instance, it will be
        used as-is for the backend filter.

        Otherwise :meth:`make_filter()` will be called to obtain the
        backend filter.  The ``filterinfo`` will be passed along to
        that call; if it is empty then ``key`` will be used instead.

        See also :meth:`remove_filter()`.  Backend filters are tracked
        via :attr:`filters`.
        """
        filtr = None

        if filterinfo and callable(filterinfo):
            # filtr = filterinfo
            raise NotImplementedError
        else:
            kwargs['key'] = key
            kwargs.setdefault('label', self.get_label(key))
            filtr = self.make_filter(filterinfo or key, **kwargs)

        self.filters[key] = filtr

    def remove_filter(self, key):
        """
        Remove the backend filter for a column.

        This removes the filter *instance*, so there is no way to
        filter by this column unless another filter is later defined
        for it.

        See also :meth:`set_filter()`.
        """
        self.filters.pop(key, None)

    def set_filter_defaults(self, **defaults):
        """
        Set default state preferences for the grid filters.

        These preferences will affect the initial grid display, until
        user requests a different filtering method.

        Each kwarg should be named by filter key, and the value should
        be a dict of preferences for that filter.  For instance::

           grid.set_filter_defaults(name={'active': True,
                                          'verb': 'contains',
                                          'value': 'foo'},
                                    value={'active': True})

        Filter defaults are tracked via :attr:`filter_defaults`.
        """
        filter_defaults = dict(getattr(self, 'filter_defaults', {}))

        for key, values in defaults.items():
            filtr = filter_defaults.setdefault(key, {})
            filtr.update(values)

        self.filter_defaults = filter_defaults

    ##############################
    # paging methods
    ##############################

    def get_pagesize_options(self, default=None):
        """
        Returns a list of default page size options for the grid.

        It will check config but if no setting exists, will fall
        back to::

           [5, 10, 20, 50, 100, 200]

        :param default: Alternate default value to return if none is
           configured.

        This method is intended for use in the constructor.  Code can
        instead access :attr:`pagesize_options` directly.
        """
        options = self.config.get_list('wuttaweb.grids.default_pagesize_options')
        if options:
            options = [int(size) for size in options
                       if size.isdigit()]
            if options:
                return options

        return default or [5, 10, 20, 50, 100, 200]

    def get_pagesize(self, default=None):
        """
        Returns the default page size for the grid.

        It will check config but if no setting exists, will fall back
        to a value from :attr:`pagesize_options` (will return ``20`` if
        that is listed; otherwise the "first" option).

        :param default: Alternate default value to return if none is
           configured.

        This method is intended for use in the constructor.  Code can
        instead access :attr:`pagesize` directly.
        """
        size = self.config.get_int('wuttaweb.grids.default_pagesize')
        if size:
            return size

        if default:
            return default

        if 20 in self.pagesize_options:
            return 20

        return self.pagesize_options[0]

    ##############################
    # configuration methods
    ##############################

    def load_settings(self, persist=True):
        """
        Load all effective settings for the grid.

        If the request GET params (query string) contains grid
        settings, they are used; otherwise the settings are loaded
        from user session.

        .. note::

           As of now, "sorting" and "pagination" settings are the only
           type supported by this logic.  Settings for "filtering"
           coming soon...

        The overall logic for this method is as follows:

        * collect settings
        * apply settings to current grid
        * optionally save settings to user session

        Saving the settings to user session will allow the grid to
        remember its current settings when user refreshes the page, or
        navigates away then comes back.  Therefore normally, settings
        are saved each time they are loaded.  Note that such settings
        are wiped upon user logout.

        :param persist: Whether the collected settings should be saved
           to the user session.
        """

        # initial default settings
        settings = {}
        if self.filterable:
            for filtr in self.filters.values():
                defaults = self.filter_defaults.get(filtr.key, {})
                settings[f'filter.{filtr.key}.active'] = defaults.get('active',
                                                                      filtr.default_active)
                settings[f'filter.{filtr.key}.verb'] = defaults.get('verb',
                                                                    filtr.get_default_verb())
                settings[f'filter.{filtr.key}.value'] = defaults.get('value',
                                                                     filtr.default_value)
        if self.sortable:
            if self.sort_defaults:
                # nb. as of writing neither Buefy nor Oruga support a
                # multi-column *default* sort; so just use first sorter
                sortinfo = self.sort_defaults[0]
                settings['sorters.length'] = 1
                settings['sorters.1.key'] = sortinfo.sortkey
                settings['sorters.1.dir'] = sortinfo.sortdir
            else:
                settings['sorters.length'] = 0
        if self.paginated and self.paginate_on_backend:
            settings['pagesize'] = self.pagesize
            settings['page'] = self.page

        # update settings dict based on what we find in the request
        # and/or user session.  always prioritize the former.

        # nb. do not read settings if user wants a reset
        if self.request.GET.get('reset-view'):
            # at this point we only have default settings, and we want
            # to keep those *and* persist them for next time, below
            pass

        elif self.request_has_settings('filter'):
            self.update_filter_settings(settings, src='request')
            if self.request_has_settings('sort'):
                self.update_sort_settings(settings, src='request')
            else:
                self.update_sort_settings(settings, src='session')
            self.update_page_settings(settings)

        elif self.request_has_settings('sort'):
            self.update_filter_settings(settings, src='session')
            self.update_sort_settings(settings, src='request')
            self.update_page_settings(settings)

        elif self.request_has_settings('page'):
            self.update_filter_settings(settings, src='session')
            self.update_sort_settings(settings, src='session')
            self.update_page_settings(settings)

        else:
            # nothing found in request, so nothing new to save
            persist = False

            # but still should load whatever is in user session
            self.update_filter_settings(settings, src='session')
            self.update_sort_settings(settings, src='session')
            self.update_page_settings(settings)

        # maybe save settings in user session, for next time
        if persist:
            self.persist_settings(settings, dest='session')

        # update ourself to reflect settings dict..

        # filtering
        if self.filterable:
            for filtr in self.filters.values():
                filtr.active = settings[f'filter.{filtr.key}.active']
                filtr.verb = settings[f'filter.{filtr.key}.verb'] or filtr.get_default_verb()
                filtr.value = settings[f'filter.{filtr.key}.value']

        # sorting
        if self.sortable:
            # nb. doing this for frontend sorting also
            self.active_sorters = []
            for i in range(1, settings['sorters.length'] + 1):
                self.active_sorters.append({
                    'key': settings[f'sorters.{i}.key'],
                    'dir': settings[f'sorters.{i}.dir'],
                })
                # TODO: i thought this was needed, but now idk?
                # # nb. when showing full index page (i.e. not partial)
                # # this implies we must set the default sorter for Vue
                # # component, and only single-column is allowed there.
                # if not self.request.GET.get('partial'):
                #     break

        # paging
        if self.paginated and self.paginate_on_backend:
            self.pagesize = settings['pagesize']
            self.page = settings['page']

    def request_has_settings(self, typ):
        """ """

        if typ == 'filter' and self.filterable:
            for filtr in self.filters.values():
                if filtr.key in self.request.GET:
                    return True
            if 'filter' in self.request.GET: # user may be applying empty filters
                return True

        elif typ == 'sort' and self.sortable and self.sort_on_backend:
            if 'sort1key' in self.request.GET:
                return True

        elif typ == 'page' and self.paginated and self.paginate_on_backend:
            for key in ['pagesize', 'page']:
                if key in self.request.GET:
                    return True

        return False

    def get_setting(self, settings, key, src='session', default=None,
                    normalize=lambda v: v):
        """ """

        if src == 'request':
            value = self.request.GET.get(key)
            if value is not None:
                try:
                    return normalize(value)
                except ValueError:
                    pass

        elif src == 'session':
            value = self.request.session.get(f'grid.{self.key}.{key}')
            if value is not None:
                return normalize(value)

        # if src had nothing, try default/existing settings
        value = settings.get(key)
        if value is not None:
            return normalize(value)

        # okay then, default it is
        return default

    def update_filter_settings(self, settings, src=None):
        """ """
        if not self.filterable:
            return

        for filtr in self.filters.values():
            prefix = f'filter.{filtr.key}'

            if src == 'request':
                # consider filter active if query string contains a value for it
                settings[f'{prefix}.active'] = filtr.key in self.request.GET
                settings[f'{prefix}.verb'] = self.get_setting(
                    settings, f'{filtr.key}.verb', src='request', default='')
                settings[f'{prefix}.value'] = self.get_setting(
                    settings, filtr.key, src='request', default='')

            elif src == 'session':
                settings[f'{prefix}.active'] = self.get_setting(
                    settings, f'{prefix}.active', src='session',
                    normalize=lambda v: str(v).lower() == 'true', default=False)
                settings[f'{prefix}.verb'] = self.get_setting(
                    settings, f'{prefix}.verb', src='session', default='')
                settings[f'{prefix}.value'] = self.get_setting(
                    settings, f'{prefix}.value', src='session', default='')

    def update_sort_settings(self, settings, src=None):
        """ """
        if not (self.sortable and self.sort_on_backend):
            return

        if src == 'request':
            i = 1
            while True:
                skey = f'sort{i}key'
                if skey in self.request.GET:
                    settings[f'sorters.{i}.key'] = self.get_setting(settings, skey,
                                                                    src='request')
                    settings[f'sorters.{i}.dir'] = self.get_setting(settings, f'sort{i}dir',
                                                                    src='request',
                                                                    default='asc')
                else:
                    break
                i += 1
            settings['sorters.length'] = i - 1

        elif src == 'session':
            settings['sorters.length'] = self.get_setting(settings, 'sorters.length',
                                                          src='session', normalize=int)
            for i in range(1, settings['sorters.length'] + 1):
                for key in ('key', 'dir'):
                    skey = f'sorters.{i}.{key}'
                    settings[skey] = self.get_setting(settings, skey, src='session')

    def update_page_settings(self, settings):
        """ """
        if not (self.paginated and self.paginate_on_backend):
            return

        # update the settings dict from request and/or user session

        # pagesize
        pagesize = self.request.GET.get('pagesize')
        if pagesize is not None:
            if pagesize.isdigit():
                settings['pagesize'] = int(pagesize)
        else:
            pagesize = self.request.session.get(f'grid.{self.key}.pagesize')
            if pagesize is not None:
                settings['pagesize'] = pagesize

        # page
        page = self.request.GET.get('page')
        if page is not None:
            if page.isdigit():
                settings['page'] = int(page)
        else:
            page = self.request.session.get(f'grid.{self.key}.page')
            if page is not None:
                settings['page'] = int(page)

    def persist_settings(self, settings, dest=None):
        """ """
        if dest not in ('session',):
            raise ValueError(f"invalid dest identifier: {dest}")

        # func to save a setting value to user session
        def persist(key, value=lambda k: settings.get(k)):
            assert dest == 'session'
            skey = f'grid.{self.key}.{key}'
            self.request.session[skey] = value(key)

        # filter settings
        if self.filterable:

            # always save all filters, with status
            for filtr in self.filters.values():
                persist(f'filter.{filtr.key}.active',
                        value=lambda k: 'true' if settings.get(k) else 'false')
                persist(f'filter.{filtr.key}.verb')
                persist(f'filter.{filtr.key}.value')

        # sort settings
        if self.sortable and self.sort_on_backend:

            # first must clear all sort settings from dest. this is
            # because number of sort settings will vary, so we delete
            # all and then write all

            if dest == 'session':
                # remove sort settings from user session
                prefix = f'grid.{self.key}.sorters.'
                for key in list(self.request.session):
                    if key.startswith(prefix):
                        del self.request.session[key]

            # now save sort settings to dest
            if 'sorters.length' in settings:
                persist('sorters.length')
                for i in range(1, settings['sorters.length'] + 1):
                    persist(f'sorters.{i}.key')
                    persist(f'sorters.{i}.dir')

        # pagination settings
        if self.paginated and self.paginate_on_backend:

            # save to dest
            persist('pagesize')
            persist('page')

    ##############################
    # data methods
    ##############################

    def get_visible_data(self):
        """
        Returns the "effective" visible data for the grid.

        This uses :attr:`data` as the starting point but may morph it
        for pagination etc. per the grid settings.

        Code can either access :attr:`data` directly, or call this
        method to get only the data for current view (e.g. assuming
        pagination is used), depending on the need.

        See also these methods which may be called by this one:

        * :meth:`filter_data()`
        * :meth:`sort_data()`
        * :meth:`paginate_data()`
        """
        data = self.data or []
        self.joined = set()

        if self.filterable:
            data = self.filter_data(data)

        if self.sortable and self.sort_on_backend:
            data = self.sort_data(data)

        if self.paginated and self.paginate_on_backend:
            self.pager = self.paginate_data(data)
            data = self.pager

        return data

    @property
    def active_filters(self):
        """
        Returns the list of currently active filters.

        This inspects each :class:`~wuttaweb.grids.filters.GridFilter`
        in :attr:`filters` and only returns the ones marked active.
        """
        return [filtr for filtr in self.filters.values()
                if filtr.active]

    def filter_data(self, data, filters=None):
        """
        Filter the given data and return the result.  This is called
        by :meth:`get_visible_data()`.

        :param filters: Optional list of filters to use.  If not
           specified, the grid's :attr:`active_filters` are used.
        """
        if filters is None:
            filters = self.active_filters
        if not filters:
            return data

        for filtr in filters:
            key = filtr.key

            if key in self.joiners and key not in self.joined:
                data = self.joiners[key](data)
                self.joined.add(key)

            try:
                data = filtr.apply_filter(data)
            except VerbNotSupported as error:
                log.warning("verb not supported for '%s' filter: %s", key, error.verb)
            except:
                log.exception("filtering data by '%s' failed!", key)

        return data

    def sort_data(self, data, sorters=None):
        """
        Sort the given data and return the result.  This is called by
        :meth:`get_visible_data()`.

        :param sorters: Optional list of sorters to use.  If not
           specified, the grid's :attr:`active_sorters` are used.
        """
        if sorters is None:
            sorters = self.active_sorters
        if not sorters:
            return data

        # nb. when data is a query, we want to apply sorters in the
        # requested order, so the final query has order_by() in the
        # correct "as-is" sequence.  however when data is a list we
        # must do the opposite, applying in the reverse order, so the
        # final list has the most "important" sort(s) applied last.
        if not isinstance(data, orm.Query):
            sorters = reversed(sorters)

        for sorter in sorters:
            sortkey = sorter['key']
            sortdir = sorter['dir']

            # cannot sort unless we have a sorter callable
            sortfunc = self.sorters.get(sortkey)
            if not sortfunc:
                return data

            # join appropriate model if needed
            if sortkey in self.joiners and sortkey not in self.joined:
                data = self.joiners[sortkey](data)
                self.joined.add(sortkey)

            # invoke the sorter
            data = sortfunc(data, sortdir)

        return data

    def paginate_data(self, data):
        """
        Apply pagination to the given data set, based on grid settings.

        This returns a "pager" object which can then be used as a
        "data replacement" in subsequent logic.

        This method is called by :meth:`get_visible_data()`.
        """
        if isinstance(data, orm.Query):
            pager = SqlalchemyOrmPage(data,
                                      items_per_page=self.pagesize,
                                      page=self.page)

        else:
            pager = paginate.Page(data,
                                  items_per_page=self.pagesize,
                                  page=self.page)

        # pager may have detected that our current page is outside the
        # valid range.  if so we should update ourself to match
        if pager.page != self.page:
            self.page = pager.page
            key = f'grid.{self.key}.page'
            if key in self.request.session:
                self.request.session[key] = self.page

            # and re-make the pager just to be safe (?)
            pager = self.paginate_data(data)

        return pager

    ##############################
    # rendering methods
    ##############################

    def render_batch_id(self, obj, key, value):
        """
        Column renderer for batch ID values.

        This is not used automatically but you can use it explicitly::

            grid.set_renderer('foo', 'batch_id')
        """
        if value is None:
            return ""

        batch_id = int(value)
        return f'{batch_id:08d}'

    def render_boolean(self, obj, key, value):
        """
        Column renderer for boolean values.

        This calls
        :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_boolean()`
        for the return value.

        This may be used automatically per
        :meth:`set_default_renderers()` or you can use it explicitly::

            grid.set_renderer('foo', 'boolean')
        """
        return self.app.render_boolean(value)

    def render_currency(self, obj, key, value, **kwargs):
        """
        Column renderer for currency values.

        This calls
        :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_currency()`
        for the return value.

        This is not used automatically but you can use it explicitly::

            grid.set_renderer('foo', 'currency')
            grid.set_renderer('foo', 'currency', scale=4)
        """
        return self.app.render_currency(value, **kwargs)

    def render_date(self, obj, key, value):
        """
        Column renderer for :class:`python:datetime.date` values.

        This calls
        :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_date()`
        for the return value.

        This may be used automatically per
        :meth:`set_default_renderers()` or you can use it explicitly::

            grid.set_renderer('foo', 'date')
        """
        dt = getattr(obj, key)
        return self.app.render_date(dt)

    def render_datetime(self, obj, key, value):
        """
        Column renderer for :class:`python:datetime.datetime` values.

        This calls
        :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_datetime()`
        for the return value.

        This may be used automatically per
        :meth:`set_default_renderers()` or you can use it explicitly::

            grid.set_renderer('foo', 'datetime')
        """
        dt = getattr(obj, key)
        return self.app.render_datetime(dt)

    def render_quantity(self, obj, key, value):
        """
        Column renderer for quantity values.

        This calls
        :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_quantity()`
        for the return value.

        This is not used automatically but you can use it explicitly::

            grid.set_renderer('foo', 'quantity')
        """
        return self.app.render_quantity(value)

    def render_table_element(
            self,
            form=None,
            template='/grids/table_element.mako',
            **context):
        """
        Render a simple Vue table element for the grid.

        This is what you want for a "simple" grid which does require a
        unique Vue component, but can instead use the standard table
        component.

        This returns something like:

        .. code-block:: html

           <b-table :data="gridContext['mykey'].data">
             <!-- columns etc. -->
           </b-table>

        See :meth:`render_vue_template()` for a more complete variant.

        Actual output will of course depend on grid attributes,
        :attr:`key`, :attr:`columns` etc.

        :param form: Reference to the
           :class:`~wuttaweb.forms.base.Form` instance which
           "contains" this grid.  This is needed in order to ensure
           the grid data is available to the form Vue component.

        :param template: Path to Mako template which is used to render
           the output.

        .. note::

           The above example shows ``gridContext['mykey'].data`` as
           the Vue data reference.  This should "just work" if you
           provide the correct ``form`` arg and the grid is contained
           directly by that form's Vue component.

           However, this may not account for all use cases.  For now
           we wait and see what comes up, but know the dust may not
           yet be settled here.
        """

        # nb. must register data for inclusion on page template
        if form:
            form.add_grid_vue_context(self)

        # otherwise logic is the same, just different template
        return self.render_vue_template(template=template, **context)

    def render_vue_tag(self, **kwargs):
        """
        Render the Vue component tag for the grid.

        By default this simply returns:

        .. code-block:: html

           <wutta-grid></wutta-grid>

        The actual output will depend on various grid attributes, in
        particular :attr:`vue_tagname`.
        """
        return HTML.tag(self.vue_tagname, **kwargs)

    def render_vue_template(
            self,
            template='/grids/vue_template.mako',
            **context):
        """
        Render the Vue template block for the grid.

        This is what you want for a "full-featured" grid which will
        exist as its own unique Vue component on the frontend.

        This returns something like:

        .. code-block:: none

           <script type="text/x-template" id="wutta-grid-template">
             <b-table>
               <!-- columns etc. -->
             </b-table>
           </script>

           <script>
               WuttaGridData = {}
               WuttaGrid = {
                   template: 'wutta-grid-template',
               }
           </script>

        .. todo::

           Why can't Sphinx render the above code block as 'html' ?

           It acts like it can't handle a ``<script>`` tag at all?

        See :meth:`render_table_element()` for a simpler variant.

        Actual output will of course depend on grid attributes,
        :attr:`vue_tagname` and :attr:`columns` etc.

        :param template: Path to Mako template which is used to render
           the output.
        """
        context['grid'] = self
        context.setdefault('request', self.request)
        output = render(template, context)
        return HTML.literal(output)

    def render_vue_finalize(self):
        """
        Render the Vue "finalize" script for the grid.

        By default this simply returns:

        .. code-block:: html

           <script>
             WuttaGrid.data = function() { return WuttaGridData }
             Vue.component('wutta-grid', WuttaGrid)
           </script>

        The actual output may depend on various grid attributes, in
        particular :attr:`vue_tagname`.
        """
        set_data = f"{self.vue_component}.data = function() {{ return {self.vue_component}Data }}"
        make_component = f"Vue.component('{self.vue_tagname}', {self.vue_component})"
        return HTML.tag('script', c=['\n',
                                     HTML.literal(set_data),
                                     '\n',
                                     HTML.literal(make_component),
                                     '\n'])

    def get_vue_columns(self):
        """
        Returns a list of Vue-compatible column definitions.

        This uses :attr:`columns` as the basis; each definition
        returned will be a dict in this format::

           {
               'field': 'foo',
               'label': "Foo",
               'sortable': True,
               'searchable': False,
           }

        The full format is determined by Buefy; see the Column section
        in its `Table docs
        <https://buefy.org/documentation/table/#api-view>`_.

        See also :meth:`get_vue_context()`.
        """
        if not self.columns:
            raise ValueError(f"you must define columns for the grid! key = {self.key}")

        columns = []
        for name in self.columns:
            columns.append({
                'field': name,
                'label': self.get_label(name),
                'sortable': self.is_sortable(name),
                'searchable': self.is_searchable(name),
            })
        return columns

    def get_vue_active_sorters(self):
        """
        Returns a list of Vue-compatible column sorter definitions.

        The list returned is the same as :attr:`active_sorters`;
        however the format used in Vue is different.  So this method
        just "converts" them to the required format, e.g.::

           # active_sorters format
           {'key': 'name', 'dir': 'asc'}

           # get_vue_active_sorters() format
           {'field': 'name', 'order': 'asc'}

        :returns: The :attr:`active_sorters` list, converted as
           described above.
        """
        sorters = []
        for sorter in self.active_sorters:
            sorters.append({'field': sorter['key'],
                            'order': sorter['dir']})
        return sorters

    def get_vue_filters(self):
        """
        Returns a list of Vue-compatible filter definitions.

        This returns the full set of :attr:`filters` but represents
        each as a simple dict with the filter state.
        """
        filters = []
        for filtr in self.filters.values():
            filters.append({
                'key': filtr.key,
                'data_type': filtr.data_type,
                'active': filtr.active,
                'visible': filtr.active,
                'verbs': filtr.get_verbs(),
                'verb_labels': filtr.get_verb_labels(),
                'valueless_verbs': filtr.get_valueless_verbs(),
                'verb': filtr.verb,
                'value': filtr.value,
                'label': filtr.label,
            })
        return filters

    def object_to_dict(self, obj):
        """ """
        try:
            dct = dict(obj)
        except TypeError:
            dct = dict(obj.__dict__)
            dct.pop('_sa_instance_state', None)
        return dct

    def get_vue_context(self):
        """
        Returns a dict of context for the grid, for use with the Vue
        component.  This contains the following keys:

        * ``data`` - list of Vue-compatible data records
        * ``row_classes`` - dict of per-row CSS classes

        This first calls :meth:`get_visible_data()` to get the
        original data set.  Each record is converted to a dict.

        Then it calls :func:`~wuttaweb.util.make_json_safe()` to
        ensure each record can be serialized to JSON.

        Then it invokes any :attr:`renderers` which are defined, to
        obtain the "final" values for each record.

        Then it adds a URL key/value for each of the :attr:`actions`
        defined, to each record.

        Then it calls :meth:`get_row_class()` for each record.  If a
        value is returned, it is added to the ``row_classes`` dict.
        Note that this dict is keyed by "zero-based row sequence as
        string" - the Vue component expects that.

        :returns: Dict of grid data/CSS context as described above.
        """
        original_data = self.get_visible_data()

        # loop thru data
        data = []
        row_classes = {}
        for i, record in enumerate(original_data, 1):
            original_record = record

            # convert record to new dict
            record = self.object_to_dict(record)

            # make all values safe for json
            record = make_json_safe(record, warn=False)

            # customize value rendering where applicable
            for key in self.renderers:
                value = record.get(key, None)
                record[key] = self.renderers[key](original_record, key, value)

            # add action urls to each record
            for action in self.actions:
                key = f'_action_url_{action.key}'
                if key not in record:
                    url = action.get_url(original_record, i)
                    if url:
                        record[key] = url

            # set row css class if applicable
            css_class = self.get_row_class(original_record, record, i)
            if css_class:
                # nb. use *string* zero-based index, for js compat
                row_classes[str(i-1)] = css_class

            data.append(record)

        return {
            'data': data,
            'row_classes': row_classes,
        }

    def get_vue_data(self):
        """ """
        warnings.warn("grid.get_vue_data() is deprecated; "
                      "please use grid.get_vue_context() instead",
                      DeprecationWarning, stacklevel=2)
        return self.get_vue_context()['data']

    def get_row_class(self, obj, data, i):
        """
        Returns the row CSS ``class`` attribute for the given record.
        This method is called by :meth:`get_vue_context()`.

        This will inspect/invoke :attr:`row_class` and return the
        value obtained from there.

        :param obj: Reference to the original model instance.

        :param data: Dict of record data for the instance; part of the
           Vue grid data set in/from :meth:`get_vue_context()`.

        :param i: One-based sequence for this object/record (row)
           within the grid.

        :returns: String of CSS class name(s), or ``None``.
        """
        if self.row_class:
            if callable(self.row_class):
                return self.row_class(obj, data, i)
            return self.row_class

    def get_vue_pager_stats(self):
        """
        Returns a simple dict with current grid pager stats.

        This is used when :attr:`paginate_on_backend` is in effect.
        """
        pager = self.pager
        return {
            'item_count': pager.item_count,
            'items_per_page': pager.items_per_page,
            'page': pager.page,
            'page_count': pager.page_count,
            'first_item': pager.first_item,
            'last_item': pager.last_item,
        }


class GridAction:
    """
    Represents a "row action" hyperlink within a grid context.

    All such actions are displayed as a group, in a dedicated
    **Actions** column in the grid.  So each row in the grid has its
    own set of action links.

    A :class:`Grid` can have one (or zero) or more of these in its
    :attr:`~Grid.actions` list.  You can call
    :meth:`~wuttaweb.views.base.View.make_grid_action()` to add custom
    actions from within a view.

    :param request: Current :term:`request` object.

    .. note::

       Some parameters are not explicitly described above.  However
       their corresponding attributes are described below.

    .. attribute:: key

       String key for the action (e.g. ``'edit'``), unique within the
       grid.

    .. attribute:: label

       Label to be displayed for the action link.  If not set, will be
       generated from :attr:`key` by calling
       :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.make_title()`.

       See also :meth:`render_label()`.

    .. attribute:: url

       URL for the action link, if applicable.  This *can* be a simple
       string, however that will cause every row in the grid to have
       the same URL for this action.

       A better way is to specify a callable which can return a unique
       URL for each record.  The callable should expect ``(obj, i)``
       args, for instance::

          def myurl(obj, i):
              return request.route_url('widgets.view', uuid=obj.uuid)

          action = GridAction(request, 'view', url=myurl)

       See also :meth:`get_url()`.

    .. attribute:: target

       Optional ``target`` attribute for the ``<a>`` tag.

    .. attribute:: click_handler

       Optional JS click handler for the action.  This value will be
       rendered as-is within the final grid template, hence the JS
       string must be callable code.  Note that ``props.row`` will be
       available in the calling context, so a couple of examples:

       * ``deleteThisThing(props.row)``
       * ``$emit('do-something', props.row)``

    .. attribute:: icon

       Name of icon to be shown for the action link.

       See also :meth:`render_icon()`.

    .. attribute:: link_class

       Optional HTML class attribute for the action's ``<a>`` tag.
    """

    def __init__(
            self,
            request,
            key,
            label=None,
            url=None,
            target=None,
            click_handler=None,
            icon=None,
            link_class=None,
    ):
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()
        self.key = key
        self.url = url
        self.target = target
        self.click_handler = click_handler
        self.label = label or self.app.make_title(key)
        self.icon = icon or key
        self.link_class = link_class or ''

    def render_icon_and_label(self):
        """
        Render the HTML snippet for action link icon and label.

        Default logic returns the output from :meth:`render_icon()`
        and :meth:`render_label()`.
        """
        html = [
            self.render_icon(),
            self.render_label(),
        ]
        return HTML.literal(' ').join(html)

    def render_icon(self):
        """
        Render the HTML snippet for the action link icon.

        This uses :attr:`icon` to identify the named icon to be shown.
        Output is something like (here ``'trash'`` is the icon name):

        .. code-block:: html

           <i class="fas fa-trash"></i>

        See also :meth:`render_icon_and_label()`.
        """
        if self.request.use_oruga:
            return HTML.tag('o-icon', icon=self.icon)

        return HTML.tag('i', class_=f'fas fa-{self.icon}')

    def render_label(self):
        """
        Render the label text for the action link.

        Default behavior is to return :attr:`label` as-is.

        See also :meth:`render_icon_and_label()`.
        """
        return self.label

    def get_url(self, obj, i=None):
        """
        Returns the action link URL for the given object (model
        instance).

        If :attr:`url` is a simple string, it is returned as-is.

        But if :attr:`url` is a callable (which is typically the most
        useful), that will be called with the same ``(obj, i)`` args
        passed along.

        :param obj: Model instance of whatever type the parent grid is
           setup to use.

        :param i: One-based sequence for the object's row within the
           parent grid.

        See also :attr:`url`.
        """
        if callable(self.url):
            return self.url(obj, i)

        return self.url



# -*- coding: utf-8; -*-

import datetime
import os
import sys
from unittest.mock import patch, MagicMock

from wuttaweb.views import upgrades as mod
from wuttjamaican.exc import ConfigurationError
from wuttaweb.progress import get_progress_session
from wuttaweb.testing import WebTestCase


class TestUpgradeView(WebTestCase):

    def make_view(self):
        return mod.UpgradeView(self.request)

    def test_includeme(self):
        self.pyramid_config.include('wuttaweb.views.upgrades')

    def test_configure_grid(self):
        model = self.app.model
        view = self.make_view()

        # sanity / coverage check
        grid = view.make_grid(model_class=model.Upgrade)
        view.configure_grid(grid)

    def test_grid_row_class(self):
        model = self.app.model
        enum = self.app.enum
        upgrade = model.Upgrade(description="test", status=enum.UpgradeStatus.PENDING)
        data = dict(upgrade)
        view = self.make_view()

        self.assertIsNone(view.grid_row_class(upgrade, data, 1))

        upgrade.status = enum.UpgradeStatus.EXECUTING
        self.assertEqual(view.grid_row_class(upgrade, data, 1), 'has-background-warning')

        upgrade.status = enum.UpgradeStatus.SUCCESS
        self.assertIsNone(view.grid_row_class(upgrade, data, 1))

        upgrade.status = enum.UpgradeStatus.FAILURE
        self.assertEqual(view.grid_row_class(upgrade, data, 1), 'has-background-warning')

    def test_configure_form(self):
        self.pyramid_config.add_route('upgrades.download', '/upgrades/{uuid}/download')
        model = self.app.model
        enum = self.app.enum
        user = model.User(username='barney')
        self.session.add(user)
        upgrade = model.Upgrade(description='test', created_by=user,
                                status=enum.UpgradeStatus.PENDING)
        self.session.add(upgrade)
        self.session.commit()
        view = self.make_view()

        # some fields exist when viewing
        with patch.object(view, 'viewing', new=True):
            form = view.make_form(model_class=model.Upgrade, model_instance=upgrade)
            self.assertIn('created', form)
            view.configure_form(form)
            self.assertIn('created', form)

        # but then are removed when creating
        with patch.object(view, 'creating', new=True):
            form = view.make_form(model_class=model.Upgrade)
            self.assertIn('created', form)
            view.configure_form(form)
            self.assertNotIn('created', form)

        # test executed, stdout/stderr when viewing
        with patch.object(view, 'viewing', new=True):

            # executed is *not* shown by default
            form = view.make_form(model_class=model.Upgrade, model_instance=upgrade)
            self.assertIn('executed', form)
            view.configure_form(form)
            self.assertNotIn('executed', form)
            self.assertNotIn('stdout_file', form)
            self.assertNotIn('stderr_file', form)

            # but it *is* shown if upgrade is executed
            upgrade.executed = datetime.datetime.now()
            upgrade.status = enum.UpgradeStatus.SUCCESS
            form = view.make_form(model_class=model.Upgrade, model_instance=upgrade)
            self.assertIn('executed', form)
            view.configure_form(form)
            self.assertIn('executed', form)
            self.assertIn('stdout_file', form)
            self.assertIn('stderr_file', form)

    def test_objectify(self):
        model = self.app.model
        enum = self.app.enum
        user = model.User(username='barney')
        self.session.add(user)
        self.session.commit()
        view = self.make_view()

        # user and status are auto-set when creating
        self.request.user = user
        self.request.method = 'POST'
        self.request.POST = {'description': "new one"}
        with patch.object(view, 'creating', new=True):
            form = view.make_model_form()
            self.assertTrue(form.validate())
            upgrade = view.objectify(form)
            self.assertEqual(upgrade.description, "new one")
            self.assertIs(upgrade.created_by, user)
            self.assertEqual(upgrade.status, enum.UpgradeStatus.PENDING)

    def test_download_path(self):
        model = self.app.model
        enum = self.app.enum

        appdir = self.mkdir('app')
        self.config.setdefault('wutta.appdir', appdir)
        self.assertEqual(self.app.get_appdir(), appdir)

        user = model.User(username='barney')
        upgrade = model.Upgrade(description='test', created_by=user,
                                status=enum.UpgradeStatus.PENDING)
        self.session.add(upgrade)
        self.session.commit()

        view = self.make_view()
        uuid = str(upgrade.uuid)

        # no filename
        path = view.download_path(upgrade, None)
        self.assertIsNone(path)

        # with filename
        path = view.download_path(upgrade, 'foo.txt')
        self.assertEqual(path, os.path.join(appdir, 'data', 'upgrades',
                                            uuid[:2], uuid[2:], 'foo.txt'))

    def test_get_upgrade_filepath(self):
        model = self.app.model
        enum = self.app.enum

        appdir = self.mkdir('app')
        self.config.setdefault('wutta.appdir', appdir)
        self.assertEqual(self.app.get_appdir(), appdir)

        user = model.User(username='barney')
        upgrade = model.Upgrade(description='test', created_by=user,
                                status=enum.UpgradeStatus.PENDING)
        self.session.add(upgrade)
        self.session.commit()

        view = self.make_view()
        uuid = str(upgrade.uuid)

        # no filename
        path = view.get_upgrade_filepath(upgrade)
        self.assertEqual(path, os.path.join(appdir, 'data', 'upgrades',
                                            uuid[:2], uuid[2:]))

        # with filename
        path = view.get_upgrade_filepath(upgrade, 'foo.txt')
        self.assertEqual(path, os.path.join(appdir, 'data', 'upgrades',
                                            uuid[:2], uuid[2:], 'foo.txt'))

    def test_delete_instance(self):
        model = self.app.model
        enum = self.app.enum

        appdir = self.mkdir('app')
        self.config.setdefault('wutta.appdir', appdir)
        self.assertEqual(self.app.get_appdir(), appdir)

        user = model.User(username='barney')
        upgrade = model.Upgrade(description='test', created_by=user,
                                status=enum.UpgradeStatus.PENDING)
        self.session.add(upgrade)
        self.session.commit()

        view = self.make_view()

        # mock stdout/stderr files
        upgrade_dir = view.get_upgrade_filepath(upgrade)
        stdout = view.get_upgrade_filepath(upgrade, 'stdout.log')
        with open(stdout, 'w') as f:
            f.write('stdout')
        stderr = view.get_upgrade_filepath(upgrade, 'stderr.log')
        with open(stderr, 'w') as f:
            f.write('stderr')

        # both upgrade and files are deleted
        self.assertTrue(os.path.exists(upgrade_dir))
        self.assertTrue(os.path.exists(stdout))
        self.assertTrue(os.path.exists(stderr))
        self.assertEqual(self.session.query(model.Upgrade).count(), 1)
        with patch.object(view, 'Session', return_value=self.session):
            view.delete_instance(upgrade)
        self.assertFalse(os.path.exists(upgrade_dir))
        self.assertFalse(os.path.exists(stdout))
        self.assertFalse(os.path.exists(stderr))
        self.assertEqual(self.session.query(model.Upgrade).count(), 0)

    def test_execute_instance(self):
        model = self.app.model
        enum = self.app.enum

        appdir = self.mkdir('app')
        self.config.setdefault('wutta.appdir', appdir)
        self.assertEqual(self.app.get_appdir(), appdir)

        user = model.User(username='barney')
        upgrade = model.Upgrade(description='test', created_by=user,
                                status=enum.UpgradeStatus.PENDING)
        self.session.add(upgrade)
        self.session.commit()

        view = self.make_view()
        self.request.user = user
        python = sys.executable

        # script not yet confiugred
        self.assertRaises(ConfigurationError, view.execute_instance, upgrade, user)

        # script w/ success
        goodpy = self.write_file('good.py', """
import sys
sys.stdout.write('hello from good.py')
sys.exit(0)
""")
        self.app.save_setting(self.session, 'wutta.upgrades.command', f'{python} {goodpy}')
        self.assertIsNone(upgrade.executed)
        self.assertIsNone(upgrade.executed_by)
        self.assertEqual(upgrade.status, enum.UpgradeStatus.PENDING)
        with patch.object(view, 'Session', return_value=self.session):
            with patch.object(self.config, 'usedb', new=True):
                view.execute_instance(upgrade, user)
        self.assertIsNotNone(upgrade.executed)
        self.assertIs(upgrade.executed_by, user)
        self.assertEqual(upgrade.status, enum.UpgradeStatus.SUCCESS)
        with open(view.get_upgrade_filepath(upgrade, 'stdout.log')) as f:
            self.assertEqual(f.read(), 'hello from good.py')
        with open(view.get_upgrade_filepath(upgrade, 'stderr.log')) as f:
            self.assertEqual(f.read(), '')

        # need a new record for next test
        upgrade = model.Upgrade(description='test', created_by=user,
                                status=enum.UpgradeStatus.PENDING)
        self.session.add(upgrade)
        self.session.commit()

        # script w/ failure
        badpy = self.write_file('bad.py', """
import sys
sys.stderr.write('hello from bad.py')
sys.exit(42)
""")
        self.app.save_setting(self.session, 'wutta.upgrades.command', f'{python} {badpy}')
        self.assertIsNone(upgrade.executed)
        self.assertIsNone(upgrade.executed_by)
        self.assertEqual(upgrade.status, enum.UpgradeStatus.PENDING)
        with patch.object(view, 'Session', return_value=self.session):
            with patch.object(self.config, 'usedb', new=True):
                view.execute_instance(upgrade, user)
        self.assertIsNotNone(upgrade.executed)
        self.assertIs(upgrade.executed_by, user)
        self.assertEqual(upgrade.status, enum.UpgradeStatus.FAILURE)
        with open(view.get_upgrade_filepath(upgrade, 'stdout.log')) as f:
            self.assertEqual(f.read(), '')
        with open(view.get_upgrade_filepath(upgrade, 'stderr.log')) as f:
            self.assertEqual(f.read(), 'hello from bad.py')

    def test_execute_progress(self):
        model = self.app.model
        enum = self.app.enum
        view = self.make_view()

        user = model.User(username='barney')
        self.session.add(user)
        upgrade = model.Upgrade(description='test', created_by=user,
                                status=enum.UpgradeStatus.PENDING)
        self.session.add(upgrade)
        self.session.commit()

        stdout = self.write_file('stdout.log', 'hello 001\n')

        self.request.matchdict = {'uuid': upgrade.uuid}
        with patch.multiple(mod.UpgradeView,
                            Session=MagicMock(return_value=self.session),
                            get_upgrade_filepath=MagicMock(return_value=stdout)):

            # nb. this is used to identify progress tracker
            self.request.session.id = 'mockid#1'

            # first call should get the full contents
            context = view.execute_progress()
            self.assertFalse(context.get('complete'))
            self.assertFalse(context.get('error'))
            # nb. newline is converted to <br>
            self.assertEqual(context['stdout'], 'hello 001<br />')

            # next call should get any new contents
            with open(stdout, 'a') as f:
                f.write('hello 002\n')
            context = view.execute_progress()
            self.assertFalse(context.get('complete'))
            self.assertFalse(context.get('error'))
            self.assertEqual(context['stdout'], 'hello 002<br />')

            # nb. switch to a different progress tracker
            self.request.session.id = 'mockid#2'

            # first call should get the full contents
            context = view.execute_progress()
            self.assertFalse(context.get('complete'))
            self.assertFalse(context.get('error'))
            self.assertEqual(context['stdout'], 'hello 001<br />hello 002<br />')

            # mark progress complete
            session = get_progress_session(self.request, 'upgrades.execute')
            session.load()
            session['complete'] = True
            session['success_msg'] = 'yay!'
            session.save()

            # next call should reflect that
            self.assertEqual(self.request.session.pop_flash(), [])
            context = view.execute_progress()
            self.assertTrue(context.get('complete'))
            self.assertFalse(context.get('error'))
            # nb. this is missing b/c we already got all contents
            self.assertNotIn('stdout', context)
            self.assertEqual(self.request.session.pop_flash(), ['yay!'])

            # nb. switch to a different progress tracker
            self.request.session.id = 'mockid#3'

            # first call should get the full contents
            context = view.execute_progress()
            self.assertFalse(context.get('complete'))
            self.assertFalse(context.get('error'))
            self.assertEqual(context['stdout'], 'hello 001<br />hello 002<br />')

            # mark progress error
            session = get_progress_session(self.request, 'upgrades.execute')
            session.load()
            session['error'] = True
            session['error_msg'] = 'omg!'
            session.save()

            # next call should reflect that
            self.assertEqual(self.request.session.pop_flash('error'), [])
            context = view.execute_progress()
            self.assertFalse(context.get('complete'))
            self.assertTrue(context.get('error'))
            # nb. this is missing b/c we already got all contents
            self.assertNotIn('stdout', context)
            self.assertEqual(self.request.session.pop_flash('error'), ['omg!'])

    def test_configure_get_simple_settings(self):
        # sanity/coverage check
        view = self.make_view()
        simple = view.configure_get_simple_settings()



# -*- coding: utf-8; -*-

from unittest.mock import patch

from sqlalchemy import orm

import colander

from wuttaweb.views import users as mod
from wuttaweb.testing import WebTestCase


class TestUserView(WebTestCase):

    def make_view(self):
        return mod.UserView(self.request)

    def test_includeme(self):
        self.pyramid_config.include('wuttaweb.views.users')

    def test_get_query(self):
        view = self.make_view()
        query = view.get_query(session=self.session)
        self.assertIsInstance(query, orm.Query)

    def test_configure_grid(self):
        model = self.app.model
        view = self.make_view()
        grid = view.make_grid(model_class=model.User)
        self.assertFalse(grid.is_linked('person'))
        view.configure_grid(grid)
        self.assertTrue(grid.is_linked('person'))

    def test_grid_row_class(self):
        model = self.app.model
        user = model.User(username='barney', active=True)
        data = dict(user)
        view = self.make_view()

        self.assertIsNone(view.grid_row_class(user, data, 1))

        user.active = False
        self.assertEqual(view.grid_row_class(user, data, 1), 'has-background-warning')

    def test_is_editable(self):
        model = self.app.model
        view = self.make_view()

        # active user is editable
        user = model.User(username='barney', active=True)
        self.assertTrue(view.is_editable(user))

        # inactive also editable
        user = model.User(username='barney', active=False)
        self.assertTrue(view.is_editable(user))

        # but not if prevent_edit flag is set
        user = model.User(username='barney', prevent_edit=True)
        self.assertFalse(view.is_editable(user))

        # unless request user is root
        self.request.is_root = True
        self.assertTrue(view.is_editable(user))

    def test_configure_form(self):
        model = self.app.model
        barney = model.User(username='barney')
        self.session.add(barney)
        self.session.commit()
        view = self.make_view()

        # person is *not* required
        with patch.object(view, 'creating', new=True):
            form = view.make_form(model_class=model.User)
            self.assertIsNone(form.is_required('person'))
            view.configure_form(form)
            self.assertFalse(form.is_required('person'))

        # password removed (always, for now)
        with patch.object(view, 'viewing', new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn('password', form)
            view.configure_form(form)
            self.assertNotIn('password', form)
        with patch.object(view, 'editing', new=True):
            form = view.make_form(model_instance=barney)
            self.assertIn('password', form)
            view.configure_form(form)
            self.assertNotIn('password', form)

    def test_unique_username(self):
        model = self.app.model
        view = self.make_view()

        user = model.User(username='foo')
        self.session.add(user)
        self.session.commit()

        with patch.object(mod, 'Session', return_value=self.session):

            # invalid if same username in data
            node = colander.SchemaNode(colander.String(), name='username')
            self.assertRaises(colander.Invalid, view.unique_username, node, 'foo')

            # but not if username belongs to current user
            view.editing = True
            self.request.matchdict = {'uuid': user.uuid}
            node = colander.SchemaNode(colander.String(), name='username')
            self.assertIsNone(view.unique_username(node, 'foo'))

    def test_objectify(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        others = model.Role(name="Others")
        self.session.add(others)
        barney = model.User(username='barney')
        auth.set_user_password(barney, 'testpass')
        barney.roles.append(blokes)
        self.session.add(barney)
        self.session.commit()
        view = self.make_view()
        view.editing = True
        self.request.matchdict = {'uuid': barney.uuid}

        # sanity check, user is just in 'blokes' role
        self.session.refresh(barney)
        self.assertEqual(len(barney.roles), 1)
        self.assertEqual(barney.roles[0].name, "Blokes")

        # form can update user password
        self.assertTrue(auth.check_user_password(barney, 'testpass'))
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barney', 'set_password': 'testpass2'}
        user = view.objectify(form, session=self.session)
        self.assertIs(user, barney)
        self.assertTrue(auth.check_user_password(barney, 'testpass2'))

        # form can update user roles
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barney', 'roles': {others.uuid}}
        user = view.objectify(form, session=self.session)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 1)
        self.assertEqual(user.roles[0].name, "Others")

    def test_update_roles(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(self.session)
        authed = auth.get_role_authenticated(self.session)
        anon = auth.get_role_anonymous(self.session)
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        others = model.Role(name="Others")
        self.session.add(others)
        barney = model.User(username='barney')
        barney.roles.append(blokes)
        self.session.add(barney)
        self.session.commit()
        view = self.make_view()
        view.editing = True
        self.request.matchdict = {'uuid': barney.uuid}

        # no error if data is missing roles
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barneyx'}
        user = view.objectify(form, session=self.session)
        self.assertIs(user, barney)
        self.assertEqual(barney.username, 'barneyx')

        # sanity check, user is just in 'blokes' role
        self.session.refresh(barney)
        self.assertEqual(len(barney.roles), 1)
        self.assertEqual(barney.roles[0].name, "Blokes")

        # let's test a bunch at once to ensure:
        # - user roles are updated
        # - authed / anon roles are not added
        # - admin role not added if current user is not root
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barney',
                          'roles': {admin.uuid, authed.uuid, anon.uuid, others.uuid}}
        user = view.objectify(form, session=self.session)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 1)
        self.assertEqual(user.roles[0].name, "Others")

        # let's test a bunch at once to ensure:
        # - user roles are updated
        # - admin role is added if current user is root
        self.request.is_root = True
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barney',
                          'roles': {admin.uuid, blokes.uuid, others.uuid}}
        user = view.objectify(form, session=self.session)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 3)
        role_uuids = set([role.uuid for role in user.roles])
        self.assertEqual(role_uuids, {admin.uuid, blokes.uuid, others.uuid})

        # admin role not removed if current user is not root
        self.request.is_root = False
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barney',
                          'roles': {blokes.uuid, others.uuid}}
        user = view.objectify(form, session=self.session)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 3)

        # admin role is removed if current user is root
        self.request.is_root = True
        form = view.make_model_form(model_instance=barney)
        form.validated = {'username': 'barney',
                          'roles': {blokes.uuid, others.uuid}}
        user = view.objectify(form, session=self.session)
        self.assertIs(user, barney)
        self.assertEqual(len(user.roles), 2)



# -*- coding: utf-8; -*-

import decimal
import json
import uuid as _uuid
from unittest import TestCase
from unittest.mock import patch, MagicMock

import colander
from fanstatic import Library, Resource
from pyramid import testing

from wuttjamaican.conf import WuttaConfig
from wuttjamaican.testing import ConfigTestCase

from wuttaweb import util as mod


class TestFieldList(TestCase):

    def test_insert_before(self):
        fields = mod.FieldList(['f1', 'f2'])
        self.assertEqual(fields, ['f1', 'f2'])

        # typical
        fields.insert_before('f1', 'XXX')
        self.assertEqual(fields, ['XXX', 'f1', 'f2'])
        fields.insert_before('f2', 'YYY')
        self.assertEqual(fields, ['XXX', 'f1', 'YYY', 'f2'])

        # appends new field if reference field is invalid
        fields.insert_before('f3', 'ZZZ')
        self.assertEqual(fields, ['XXX', 'f1', 'YYY', 'f2', 'ZZZ'])

    def test_insert_after(self):
        fields = mod.FieldList(['f1', 'f2'])
        self.assertEqual(fields, ['f1', 'f2'])

        # typical
        fields.insert_after('f1', 'XXX')
        self.assertEqual(fields, ['f1', 'XXX', 'f2'])
        fields.insert_after('XXX', 'YYY')
        self.assertEqual(fields, ['f1', 'XXX', 'YYY', 'f2'])

        # appends new field if reference field is invalid
        fields.insert_after('f3', 'ZZZ')
        self.assertEqual(fields, ['f1', 'XXX', 'YYY', 'f2', 'ZZZ'])

    def test_set_sequence(self):
        fields = mod.FieldList(['f5', 'f1', 'f3', 'f4', 'f2'])

        # setting sequence will only "sort" for explicit fields.
        # other fields remain in original order, but at the end.
        fields.set_sequence(['f1', 'f2', 'f3'])
        self.assertEqual(fields, ['f1', 'f2', 'f3', 'f5', 'f4'])


class TestGetLibVer(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = testing.DummyRequest()
        self.request.wutta_config = self.config

    def test_buefy_default(self):
        version = mod.get_libver(self.request, 'buefy')
        self.assertEqual(version, '0.9.25')

    def test_buefy_custom_old(self):
        self.config.setdefault('wuttaweb.buefy_version', '0.9.29')
        version = mod.get_libver(self.request, 'buefy')
        self.assertEqual(version, '0.9.29')

    def test_buefy_custom_old_tailbone(self):
        self.config.setdefault('tailbone.libver.buefy', '0.9.28')
        version = mod.get_libver(self.request, 'buefy', prefix='tailbone')
        self.assertEqual(version, '0.9.28')

    def test_buefy_custom_new(self):
        self.config.setdefault('wuttaweb.libver.buefy', '0.9.29')
        version = mod.get_libver(self.request, 'buefy')
        self.assertEqual(version, '0.9.29')

    def test_buefy_configured_only(self):
        version = mod.get_libver(self.request, 'buefy', configured_only=True)
        self.assertIsNone(version)

    def test_buefy_default_only(self):
        self.config.setdefault('wuttaweb.libver.buefy', '0.9.29')
        version = mod.get_libver(self.request, 'buefy', default_only=True)
        self.assertEqual(version, '0.9.25')

    def test_buefy_css_default(self):
        version = mod.get_libver(self.request, 'buefy.css')
        self.assertEqual(version, '0.9.25')

    def test_buefy_css_custom_old(self):
        # nb. this uses same setting as buefy (js)
        self.config.setdefault('wuttaweb.buefy_version', '0.9.29')
        version = mod.get_libver(self.request, 'buefy.css')
        self.assertEqual(version, '0.9.29')

    def test_buefy_css_custom_new(self):
        # nb. this uses same setting as buefy (js)
        self.config.setdefault('wuttaweb.libver.buefy', '0.9.29')
        version = mod.get_libver(self.request, 'buefy.css')
        self.assertEqual(version, '0.9.29')

    def test_buefy_css_configured_only(self):
        version = mod.get_libver(self.request, 'buefy.css', configured_only=True)
        self.assertIsNone(version)

    def test_buefy_css_default_only(self):
        self.config.setdefault('wuttaweb.libver.buefy', '0.9.29')
        version = mod.get_libver(self.request, 'buefy.css', default_only=True)
        self.assertEqual(version, '0.9.25')

    def test_vue_default(self):
        version = mod.get_libver(self.request, 'vue')
        self.assertEqual(version, '2.6.14')

    def test_vue_custom_old(self):
        self.config.setdefault('wuttaweb.vue_version', '3.4.31')
        version = mod.get_libver(self.request, 'vue')
        self.assertEqual(version, '3.4.31')

    def test_vue_custom_new(self):
        self.config.setdefault('wuttaweb.libver.vue', '3.4.31')
        version = mod.get_libver(self.request, 'vue')
        self.assertEqual(version, '3.4.31')

    def test_vue_configured_only(self):
        version = mod.get_libver(self.request, 'vue', configured_only=True)
        self.assertIsNone(version)

    def test_vue_default_only(self):
        self.config.setdefault('wuttaweb.libver.vue', '3.4.31')
        version = mod.get_libver(self.request, 'vue', default_only=True)
        self.assertEqual(version, '2.6.14')

    def test_vue_resource_default(self):
        version = mod.get_libver(self.request, 'vue_resource')
        self.assertEqual(version, '1.5.3')

    def test_vue_resource_custom(self):
        self.config.setdefault('wuttaweb.libver.vue_resource', '1.5.3')
        version = mod.get_libver(self.request, 'vue_resource')
        self.assertEqual(version, '1.5.3')

    def test_fontawesome_default(self):
        version = mod.get_libver(self.request, 'fontawesome')
        self.assertEqual(version, '5.3.1')

    def test_fontawesome_custom(self):
        self.config.setdefault('wuttaweb.libver.fontawesome', '5.6.3')
        version = mod.get_libver(self.request, 'fontawesome')
        self.assertEqual(version, '5.6.3')

    def test_bb_vue_default(self):
        version = mod.get_libver(self.request, 'bb_vue')
        self.assertEqual(version, '3.4.31')

    def test_bb_vue_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_vue', '3.4.30')
        version = mod.get_libver(self.request, 'bb_vue')
        self.assertEqual(version, '3.4.30')

    def test_bb_oruga_default(self):
        version = mod.get_libver(self.request, 'bb_oruga')
        self.assertEqual(version, '0.8.12')

    def test_bb_oruga_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_oruga', '0.8.11')
        version = mod.get_libver(self.request, 'bb_oruga')
        self.assertEqual(version, '0.8.11')

    def test_bb_oruga_bulma_default(self):
        version = mod.get_libver(self.request, 'bb_oruga_bulma')
        self.assertEqual(version, '0.3.0')
        version = mod.get_libver(self.request, 'bb_oruga_bulma_css')
        self.assertEqual(version, '0.3.0')

    def test_bb_oruga_bulma_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_oruga_bulma', '0.2.11')
        version = mod.get_libver(self.request, 'bb_oruga_bulma')
        self.assertEqual(version, '0.2.11')

    def test_bb_fontawesome_svg_core_default(self):
        version = mod.get_libver(self.request, 'bb_fontawesome_svg_core')
        self.assertEqual(version, '6.5.2')

    def test_bb_fontawesome_svg_core_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_fontawesome_svg_core', '6.5.1')
        version = mod.get_libver(self.request, 'bb_fontawesome_svg_core')
        self.assertEqual(version, '6.5.1')

    def test_bb_free_solid_svg_icons_default(self):
        version = mod.get_libver(self.request, 'bb_free_solid_svg_icons')
        self.assertEqual(version, '6.5.2')

    def test_bb_free_solid_svg_icons_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_free_solid_svg_icons', '6.5.1')
        version = mod.get_libver(self.request, 'bb_free_solid_svg_icons')
        self.assertEqual(version, '6.5.1')

    def test_bb_vue_fontawesome_default(self):
        version = mod.get_libver(self.request, 'bb_vue_fontawesome')
        self.assertEqual(version, '3.0.6')

    def test_bb_vue_fontawesome_custom(self):
        self.config.setdefault('wuttaweb.libver.bb_vue_fontawesome', '3.0.8')
        version = mod.get_libver(self.request, 'bb_vue_fontawesome')
        self.assertEqual(version, '3.0.8')


libcache = Library('testing', 'libcache')
vue_js = Resource(libcache, 'vue.js')
vue_resource_js = Resource(libcache, 'vue_resource.js')
buefy_js = Resource(libcache, 'buefy.js')
buefy_css = Resource(libcache, 'buefy.css')
fontawesome_js = Resource(libcache, 'fontawesome.js')
bb_vue_js = Resource(libcache, 'bb_vue.js')
bb_oruga_js = Resource(libcache, 'bb_oruga.js')
bb_oruga_bulma_js = Resource(libcache, 'bb_oruga_bulma.js')
bb_oruga_bulma_css = Resource(libcache, 'bb_oruga_bulma.css')
bb_fontawesome_svg_core_js = Resource(libcache, 'bb_fontawesome_svg_core.js')
bb_free_solid_svg_icons_js = Resource(libcache, 'bb_free_solid_svg_icons.js')
bb_vue_fontawesome_js = Resource(libcache, 'bb_vue_fontawesome.js')


class TestGetLibUrl(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = testing.DummyRequest(wutta_config=self.config)
        self.pyramid_config = testing.setUp(request=self.request)

    def tearDown(self):
        testing.tearDown()

    def setup_fanstatic(self, register=True):
        self.pyramid_config.include('pyramid_fanstatic')
        if register:
            self.config.setdefault('wuttaweb.static_libcache.module',
                                   'tests.test_util')

        needed = MagicMock()
        needed.library_url = MagicMock(return_value='/fanstatic')
        self.request.environ['fanstatic.needed'] = needed
        self.request.script_name = '/wutta'

    def test_buefy_default(self):
        url = mod.get_liburl(self.request, 'buefy')
        self.assertEqual(url, 'https://unpkg.com/buefy@0.9.25/dist/buefy.min.js')

    def test_buefy_custom(self):
        self.config.setdefault('wuttaweb.liburl.buefy', '/lib/buefy.js')
        url = mod.get_liburl(self.request, 'buefy')
        self.assertEqual(url, '/lib/buefy.js')

    def test_buefy_custom_tailbone(self):
        self.config.setdefault('tailbone.liburl.buefy', '/tailbone/buefy.js')
        url = mod.get_liburl(self.request, 'buefy', prefix='tailbone')
        self.assertEqual(url, '/tailbone/buefy.js')

    def test_buefy_default_only(self):
        self.config.setdefault('wuttaweb.liburl.buefy', '/lib/buefy.js')
        url = mod.get_liburl(self.request, 'buefy', default_only=True)
        self.assertEqual(url, 'https://unpkg.com/buefy@0.9.25/dist/buefy.min.js')

    def test_buefy_configured_only(self):
        url = mod.get_liburl(self.request, 'buefy', configured_only=True)
        self.assertIsNone(url)

    def test_buefy_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'buefy')
        self.assertEqual(url, '/wutta/fanstatic/buefy.js')

    def test_buefy_fanstatic_tailbone(self):
        self.setup_fanstatic(register=False)
        self.config.setdefault('tailbone.static_libcache.module', 'tests.test_util')
        url = mod.get_liburl(self.request, 'buefy', prefix='tailbone')
        self.assertEqual(url, '/wutta/fanstatic/buefy.js')

    def test_buefy_css_default(self):
        url = mod.get_liburl(self.request, 'buefy.css')
        self.assertEqual(url, 'https://unpkg.com/buefy@0.9.25/dist/buefy.min.css')

    def test_buefy_css_custom(self):
        self.config.setdefault('wuttaweb.liburl.buefy.css', '/lib/buefy.css')
        url = mod.get_liburl(self.request, 'buefy.css')
        self.assertEqual(url, '/lib/buefy.css')

    def test_buefy_css_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'buefy.css')
        self.assertEqual(url, '/wutta/fanstatic/buefy.css')

    def test_vue_default(self):
        url = mod.get_liburl(self.request, 'vue')
        self.assertEqual(url, 'https://unpkg.com/vue@2.6.14/dist/vue.min.js')

    def test_vue_custom(self):
        self.config.setdefault('wuttaweb.liburl.vue', '/lib/vue.js')
        url = mod.get_liburl(self.request, 'vue')
        self.assertEqual(url, '/lib/vue.js')

    def test_vue_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'vue')
        self.assertEqual(url, '/wutta/fanstatic/vue.js')

    def test_vue_resource_default(self):
        url = mod.get_liburl(self.request, 'vue_resource')
        self.assertEqual(url, 'https://cdn.jsdelivr.net/npm/vue-resource@1.5.3')

    def test_vue_resource_custom(self):
        self.config.setdefault('wuttaweb.liburl.vue_resource', '/lib/vue-resource.js')
        url = mod.get_liburl(self.request, 'vue_resource')
        self.assertEqual(url, '/lib/vue-resource.js')

    def test_vue_resource_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'vue_resource')
        self.assertEqual(url, '/wutta/fanstatic/vue_resource.js')

    def test_fontawesome_default(self):
        url = mod.get_liburl(self.request, 'fontawesome')
        self.assertEqual(url, 'https://use.fontawesome.com/releases/v5.3.1/js/all.js')

    def test_fontawesome_custom(self):
        self.config.setdefault('wuttaweb.liburl.fontawesome', '/lib/fontawesome.js')
        url = mod.get_liburl(self.request, 'fontawesome')
        self.assertEqual(url, '/lib/fontawesome.js')

    def test_fontawesome_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'fontawesome')
        self.assertEqual(url, '/wutta/fanstatic/fontawesome.js')

    def test_bb_vue_default(self):
        url = mod.get_liburl(self.request, 'bb_vue')
        self.assertEqual(url, 'https://unpkg.com/vue@3.4.31/dist/vue.esm-browser.prod.js')

    def test_bb_vue_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_vue', '/lib/vue.js')
        url = mod.get_liburl(self.request, 'bb_vue')
        self.assertEqual(url, '/lib/vue.js')

    def test_bb_vue_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'bb_vue')
        self.assertEqual(url, '/wutta/fanstatic/bb_vue.js')

    def test_bb_oruga_default(self):
        url = mod.get_liburl(self.request, 'bb_oruga')
        self.assertEqual(url, 'https://unpkg.com/@oruga-ui/oruga-next@0.8.12/dist/oruga.mjs')

    def test_bb_oruga_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_oruga', '/lib/oruga.js')
        url = mod.get_liburl(self.request, 'bb_oruga')
        self.assertEqual(url, '/lib/oruga.js')

    def test_bb_oruga_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'bb_oruga')
        self.assertEqual(url, '/wutta/fanstatic/bb_oruga.js')

    def test_bb_oruga_bulma_default(self):
        url = mod.get_liburl(self.request, 'bb_oruga_bulma')
        self.assertEqual(url, 'https://unpkg.com/@oruga-ui/theme-bulma@0.3.0/dist/bulma.mjs')

    def test_bb_oruga_bulma_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_oruga_bulma', '/lib/oruga_bulma.js')
        url = mod.get_liburl(self.request, 'bb_oruga_bulma')
        self.assertEqual(url, '/lib/oruga_bulma.js')

    def test_bb_oruga_bulma_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'bb_oruga_bulma')
        self.assertEqual(url, '/wutta/fanstatic/bb_oruga_bulma.js')

    def test_bb_oruga_bulma_css_default(self):
        url = mod.get_liburl(self.request, 'bb_oruga_bulma_css')
        self.assertEqual(url, 'https://unpkg.com/@oruga-ui/theme-bulma@0.3.0/dist/bulma.css')

    def test_bb_oruga_bulma_css_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_oruga_bulma_css', '/lib/oruga-bulma.css')
        url = mod.get_liburl(self.request, 'bb_oruga_bulma_css')
        self.assertEqual(url, '/lib/oruga-bulma.css')

    def test_bb_oruga_bulma_css_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'bb_oruga_bulma_css')
        self.assertEqual(url, '/wutta/fanstatic/bb_oruga_bulma.css')

    def test_bb_fontawesome_svg_core_default(self):
        url = mod.get_liburl(self.request, 'bb_fontawesome_svg_core')
        self.assertEqual(url, 'https://cdn.jsdelivr.net/npm/@fortawesome/fontawesome-svg-core@6.5.2/+esm')

    def test_bb_fontawesome_svg_core_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_fontawesome_svg_core', '/lib/fontawesome-svg-core.js')
        url = mod.get_liburl(self.request, 'bb_fontawesome_svg_core')
        self.assertEqual(url, '/lib/fontawesome-svg-core.js')

    def test_bb_fontawesome_svg_core_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'bb_fontawesome_svg_core')
        self.assertEqual(url, '/wutta/fanstatic/bb_fontawesome_svg_core.js')

    def test_bb_free_solid_svg_icons_default(self):
        url = mod.get_liburl(self.request, 'bb_free_solid_svg_icons')
        self.assertEqual(url, 'https://cdn.jsdelivr.net/npm/@fortawesome/free-solid-svg-icons@6.5.2/+esm')

    def test_bb_free_solid_svg_icons_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_free_solid_svg_icons', '/lib/free-solid-svg-icons.js')
        url = mod.get_liburl(self.request, 'bb_free_solid_svg_icons')
        self.assertEqual(url, '/lib/free-solid-svg-icons.js')

    def test_bb_free_solid_svg_icons_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'bb_free_solid_svg_icons')
        self.assertEqual(url, '/wutta/fanstatic/bb_free_solid_svg_icons.js')

    def test_bb_vue_fontawesome_default(self):
        url = mod.get_liburl(self.request, 'bb_vue_fontawesome')
        self.assertEqual(url, 'https://cdn.jsdelivr.net/npm/@fortawesome/vue-fontawesome@3.0.6/+esm')

    def test_bb_vue_fontawesome_custom(self):
        self.config.setdefault('wuttaweb.liburl.bb_vue_fontawesome', '/lib/vue-fontawesome.js')
        url = mod.get_liburl(self.request, 'bb_vue_fontawesome')
        self.assertEqual(url, '/lib/vue-fontawesome.js')

    def test_bb_vue_fontawesome_fanstatic(self):
        self.setup_fanstatic()
        url = mod.get_liburl(self.request, 'bb_vue_fontawesome')
        self.assertEqual(url, '/wutta/fanstatic/bb_vue_fontawesome.js')


class TestGetFormData(TestCase):

    def setUp(self):
        self.config = WuttaConfig()

    def make_request(self, **kwargs):
        kwargs.setdefault('wutta_config', self.config)
        kwargs.setdefault('POST', {'foo1': 'bar'})
        kwargs.setdefault('json_body', {'foo2': 'baz'})
        return testing.DummyRequest(**kwargs)

    def test_default(self):
        request = self.make_request()
        data = mod.get_form_data(request)
        self.assertEqual(data, {'foo1': 'bar'})

    def test_is_xhr(self):
        request = self.make_request(POST=None, is_xhr=True)
        data = mod.get_form_data(request)
        self.assertEqual(data, {'foo2': 'baz'})

    def test_content_type(self):
        request = self.make_request(POST=None, content_type='application/json')
        data = mod.get_form_data(request)
        self.assertEqual(data, {'foo2': 'baz'})


class TestGetModelFields(ConfigTestCase):

    def test_empty_model_class(self):
        fields = mod.get_model_fields(self.config, None)
        self.assertIsNone(fields)

    def test_unknown_model_class(self):
        fields = mod.get_model_fields(self.config, TestCase)
        self.assertIsNone(fields)

    def test_basic(self):
        model = self.app.model
        fields = mod.get_model_fields(self.config, model.Setting)
        self.assertEqual(fields, ['name', 'value'])

    def test_include_fk(self):
        model = self.app.model

        # fk excluded by default
        fields = mod.get_model_fields(self.config, model.User)
        self.assertNotIn('person_uuid', fields)
        self.assertIn('person', fields)

        # fk can be included
        fields = mod.get_model_fields(self.config, model.User, include_fk=True)
        self.assertIn('person_uuid', fields)
        self.assertIn('person', fields)

    def test_avoid_versions(self):
        model = self.app.model

        mapper = MagicMock(iterate_properties = [
            MagicMock(key='uuid'),
            MagicMock(key='full_name'),
            MagicMock(key='first_name'),
            MagicMock(key='middle_name'),
            MagicMock(key='last_name'),
            MagicMock(key='versions'),
        ])

        with patch.object(mod, 'sa') as sa:
            sa.inspect.return_value = mapper

            with patch.object(self.app, 'continuum_is_enabled', return_value=True):
                fields = mod.get_model_fields(self.config, model.Person)
                # nb. no versions field
                self.assertEqual(set(fields), set(['uuid', 'full_name', 'first_name',
                                                   'middle_name', 'last_name']))


class TestGetCsrfToken(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_same_token(self):

        # same token returned for same request
        # TODO: dummy request is always returning same token!
        # so this isn't really testing anything.. :(
        first = mod.get_csrf_token(self.request)
        self.assertIsNotNone(first)
        second = mod.get_csrf_token(self.request)
        self.assertEqual(first, second)

        # TODO: ideally would make a new request here and confirm it
        # gets a different token, but see note above..

    def test_new_token(self):

        # nb. dummy request always returns same token, so must
        # trick it into thinking it doesn't have one yet
        with patch.object(self.request.session, 'get_csrf_token', return_value=None):
            token = mod.get_csrf_token(self.request)
            self.assertIsNotNone(token)


class TestRenderCsrfToken(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_basics(self):
        html = mod.render_csrf_token(self.request)
        self.assertIn('type="hidden"', html)
        self.assertIn('name="_csrf"', html)
        token = mod.get_csrf_token(self.request)
        self.assertIn(f'value="{token}"', html)


class TestMakeJsonSafe(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.app = self.config.get_app()

    def test_null(self):
        value = mod.make_json_safe(colander.null)
        self.assertIsNone(value)

        value = mod.make_json_safe(None)
        self.assertIsNone(value)

    def test_invalid(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")
        self.assertRaises(TypeError, json.dumps, person)
        value = mod.make_json_safe(person, key='person')
        self.assertEqual(value, "Betty Boop")

    def test_uuid(self):
        uuid = _uuid.uuid4()
        value = mod.make_json_safe(uuid)
        self.assertEqual(value, uuid.hex)

    def test_decimal(self):
        value = decimal.Decimal('42.42')
        self.assertNotEqual(value, 42.42)
        result = mod.make_json_safe(value)
        self.assertEqual(result, 42.42)

    def test_dict(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")

        data = {
            'foo': 'bar',
            'person': person,
        }

        self.assertRaises(TypeError, json.dumps, data)
        value = mod.make_json_safe(data)
        self.assertEqual(value, {
            'foo': 'bar',
            'person': "Betty Boop",
        })

    def test_list(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")

        data = [
            'foo',
            'bar',
            person,
        ]

        self.assertRaises(TypeError, json.dumps, data)
        value = mod.make_json_safe(data)
        self.assertEqual(value, [
            'foo',
            'bar',
            "Betty Boop",
        ])



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for users
"""

import colander

from wuttjamaican.db.model import User
from wuttaweb.views import MasterView
from wuttaweb.forms import widgets
from wuttaweb.forms.schema import PersonRef, RoleRefs
from wuttaweb.db import Session


class UserView(MasterView):
    """
    Master view for users.

    Default route prefix is ``users``.

    Notable URLs provided by this class:

    * ``/users/``
    * ``/users/new``
    * ``/users/XXX``
    * ``/users/XXX/edit``
    * ``/users/XXX/delete``
    """
    model_class = User

    grid_columns = [
        'username',
        'person',
        'active',
    ]

    filter_defaults = {
        'username': {'active': True},
        'active': {'active': True, 'verb': 'is_true'},
    }
    sort_defaults = 'username'

    def get_query(self, session=None):
        """ """
        query = super().get_query(session=session)

        # nb. always join Person
        model = self.app.model
        query = query.outerjoin(model.Person)

        return query

    def configure_grid(self, g):
        """ """
        super().configure_grid(g)
        model = self.app.model

        # never show these
        g.remove('person_uuid',
                 'role_refs',
                 'password')

        # username
        g.set_link('username')

        # person
        g.set_link('person')
        g.set_sorter('person', model.Person.full_name)
        g.set_filter('person', model.Person.full_name,
                     label="Person Full Name")

    def grid_row_class(self, user, data, i):
        """ """
        if not user.active:
            return 'has-background-warning'

    def is_editable(self, user):
        """ """

        # only root can edit certain users
        if user.prevent_edit and not self.request.is_root:
            return False

        return True

    def configure_form(self, f):
        """ """
        super().configure_form(f)
        user = f.model_instance

        # never show these
        f.remove('person_uuid',
                 'role_refs')

        # person
        f.set_node('person', PersonRef(self.request, empty_option=True))
        f.set_required('person', False)

        # username
        f.set_validator('username', self.unique_username)

        # password
        # nb. we must avoid 'password' as field name since
        # ColanderAlchemy wants to handle the raw/hashed value
        f.remove('password')
        # nb. no need for password field if readonly
        if self.creating or self.editing:
            # nb. use 'set_password' as field name
            f.append('set_password')
            f.set_required('set_password', False)
            f.set_widget('set_password', widgets.CheckedPasswordWidget())

        # roles
        f.append('roles')
        f.set_node('roles', RoleRefs(self.request))
        if not self.creating:
            f.set_default('roles', [role.uuid.hex for role in user.roles])

    def unique_username(self, node, value):
        """ """
        model = self.app.model
        session = Session()

        query = session.query(model.User)\
                       .filter(model.User.username == value)

        if self.editing:
            uuid = self.request.matchdict['uuid']
            query = query.filter(model.User.uuid != uuid)

        if query.count():
            node.raise_invalid("Username must be unique")

    def objectify(self, form, session=None):
        """ """
        data = form.validated

        # normal logic first
        user = super().objectify(form)

        # maybe set user password
        if 'set_password' in form and data.get('set_password'):
            auth = self.app.get_auth_handler()
            auth.set_user_password(user, data['set_password'])

        # update roles for user
        # TODO
        # if self.has_perm('edit_roles'):
        self.update_roles(user, form, session=session)

        return user

    def update_roles(self, user, form, session=None):
        """ """
        # TODO
        # if not self.has_perm('edit_roles'):
        #     return
        data = form.validated
        if 'roles' not in data:
            return

        model = self.app.model
        session = session or Session()
        auth = self.app.get_auth_handler()

        old_roles = set([role.uuid for role in user.roles])
        new_roles = data['roles']

        admin = auth.get_role_administrator(session)
        ignored = {
            auth.get_role_authenticated(session).uuid,
            auth.get_role_anonymous(session).uuid,
        }

        # add any new roles for the user, taking care to avoid certain
        # unwanted operations for built-in roles
        for uuid in new_roles:
            if uuid in ignored:
                continue
            if uuid in old_roles:
                continue
            if uuid == admin.uuid and not self.request.is_root:
                continue
            role = session.get(model.Role, uuid)
            user.roles.append(role)

        # remove any roles which were *not* specified, taking care to
        # avoid certain unwanted operations for built-in roles
        for uuid in old_roles:
            if uuid in new_roles:
                continue
            if uuid == admin.uuid and not self.request.is_root:
                continue
            role = session.get(model.Role, uuid)
            user.roles.remove(role)

    @classmethod
    def defaults(cls, config):
        """ """

        # nb. User may come from custom model
        wutta_config = config.registry.settings['wutta_config']
        app = wutta_config.get_app()
        cls.model_class = app.model.User

        cls._defaults(config)


def defaults(config, **kwargs):
    base = globals()

    UserView = kwargs.get('UserView', base['UserView'])
    UserView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Grid Filters
"""

import datetime
import logging

import sqlalchemy as sa

from wuttjamaican.util import UNSPECIFIED


log = logging.getLogger(__name__)


class VerbNotSupported(Exception):
    """ """

    def __init__(self, verb):
        self.verb = verb

    def __str__(self):
        return f"unknown filter verb not supported: {self.verb}"


class GridFilter:
    """
    Filter option for a grid.  Represents both the "features" as well
    as "state" for the filter.

    :param request: Current :term:`request` object.

    :param model_property: Property of a model class, representing the
       column by which to filter.  For instance,
       ``model.Person.full_name``.

    :param \**kwargs: Any additional kwargs will be set as attributes
       on the filter instance.

    Filter instances have the following attributes:

    .. attribute:: key

       Unique key for the filter.  This often corresponds to a "column
       name" for the grid, but not always.

    .. attribute:: label

       Display label for the filter field.

    .. attribute:: data_type

       Simplistic "data type" which the filter supports.  So far this
       will be one of:

       * ``'string'``
       * ``'date'``

       Note that this mainly applies to the "value input" used by the
       filter.  There is no data type for boolean since it does not
       need a value input; the verb is enough.

    .. attribute:: active

       Boolean indicating whether the filter is currently active.

       See also :attr:`verb` and :attr:`value`.

    .. attribute:: verb

       Verb for current filter, if :attr:`active` is true.

       See also :attr:`value`.

    .. attribute:: value

       Value for current filter, if :attr:`active` is true.

       See also :attr:`verb`.

    .. attribute:: default_active

       Boolean indicating whether the filter should be active by
       default, i.e. when first displaying the grid.

       See also :attr:`default_verb` and :attr:`default_value`.

    .. attribute:: default_verb

       Filter verb to use by default.  This will be auto-selected when
       the filter is first activated, or when first displaying the
       grid if :attr:`default_active` is true.

       See also :attr:`default_value`.

    .. attribute:: default_value

       Filter value to use by default.  This will be auto-populated
       when the filter is first activated, or when first displaying
       the grid if :attr:`default_active` is true.

       See also :attr:`default_verb`.
    """
    data_type = 'string'
    default_verbs = ['equal', 'not_equal']

    default_verb_labels = {
        'is_any':               "is any",
        'equal':                "equal to",
        'not_equal':            "not equal to",
        'greater_than':         "greater than",
        'greater_equal':        "greater than or equal to",
        'less_than':            "less than",
        'less_equal':           "less than or equal to",
        # 'between':              "between",
        'is_true':              "is true",
        'is_false':             "is false",
        'is_false_null':        "is false or null",
        'is_null':              "is null",
        'is_not_null':          "is not null",
        'contains':             "contains",
        'does_not_contain':     "does not contain",
    }

    valueless_verbs = [
        'is_any',
        'is_true',
        'is_false',
        'is_false_null',
        'is_null',
        'is_not_null',
    ]

    def __init__(
            self,
            request,
            key,
            label=None,
            verbs=None,
            default_active=False,
            default_verb=None,
            default_value=None,
            **kwargs,
    ):
        self.request = request
        self.key = key
        self.config = self.request.wutta_config
        self.app = self.config.get_app()
        self.label = label or self.app.make_title(self.key)

        # active
        self.default_active = default_active
        self.active = self.default_active

        # verb
        if verbs is not None:
            self.verbs = verbs
        if default_verb:
            self.default_verb = default_verb

        # value
        self.default_value = default_value
        self.value = self.default_value

        self.__dict__.update(kwargs)

    def __repr__(self):
        verb = getattr(self, 'verb', None)
        return (f"{self.__class__.__name__}("
                f"key='{self.key}', "
                f"active={self.active}, "
                f"verb={repr(verb)}, "
                f"value={repr(self.value)})")

    def get_verbs(self):
        """
        Returns the list of verbs supported by the filter.
        """
        verbs = None

        if hasattr(self, 'verbs'):
            verbs = self.verbs

        else:
            verbs = self.default_verbs

        if callable(verbs):
            verbs = verbs()
        verbs = list(verbs)

        if self.nullable:
            if 'is_null' not in verbs:
                verbs.append('is_null')
            if 'is_not_null' not in verbs:
                verbs.append('is_not_null')

        if 'is_any' not in verbs:
            verbs.append('is_any')

        return verbs

    def get_verb_labels(self):
        """
        Returns a dict of all defined verb labels.
        """
        # TODO: should traverse hierarchy
        labels = dict([(verb, verb) for verb in self.get_verbs()])
        labels.update(self.default_verb_labels)
        return labels

    def get_valueless_verbs(self):
        """
        Returns a list of verb names which do not need a value.
        """
        return self.valueless_verbs

    def get_default_verb(self):
        """
        Returns the default verb for the filter.
        """
        verb = None

        if hasattr(self, 'default_verb'):
            verb = self.default_verb

        elif hasattr(self, 'verb'):
            verb = self.verb

        if not verb:
            verbs = self.get_verbs()
            if verbs:
                verb = verbs[0]

        return verb

    def apply_filter(self, data, verb=None, value=UNSPECIFIED):
        """
        Filter the given data set according to a verb/value pair.

        If verb and/or value are not specified, will use :attr:`verb`
        and/or :attr:`value` instead.

        This method does not directly filter the data; rather it
        delegates (based on ``verb``) to some other method.  The
        latter may choose *not* to filter the data, e.g. if ``value``
        is empty, in which case this may return the original data set
        unchanged.

        :returns: The (possibly) filtered data set.
        """
        if verb is None:
            verb = self.verb
        if not verb:
            verb = self.get_default_verb()
            log.warn("missing verb for '%s' filter, will use default verb: %s",
                     self.key, verb)

        # only attempt for known verbs
        if verb not in self.get_verbs():
            raise VerbNotSupported(verb)

        # fallback value
        if value is UNSPECIFIED:
            value = self.value

        # locate filter method
        func = getattr(self, f'filter_{verb}', None)
        if not func:
            raise VerbNotSupported(verb)

        # invoke filter method
        return func(data, value)

    def filter_is_any(self, data, value):
        """
        This is a no-op which always ignores the value and returns the
        data as-is.
        """
        return data


class AlchemyFilter(GridFilter):
    """
    Filter option for a grid with SQLAlchemy query data.

    This is a subclass of :class:`GridFilter`.  It requires a
    ``model_property`` to know how to filter the query.

    :param model_property: Property of a model class, representing the
       column by which to filter.  For instance,
       ``model.Person.full_name``.

    :param nullable: Boolean indicating whether the filter should
       include ``is_null`` and ``is_not_null`` verbs.  If not
       specified, the column will be inspected and use its nullable
       flag.
    """

    def __init__(self, *args, **kwargs):
        nullable = kwargs.pop('nullable', None)
        super().__init__(*args, **kwargs)

        self.nullable = nullable
        if self.nullable is None:
            columns = self.model_property.prop.columns
            if len(columns) == 1:
                self.nullable = columns[0].nullable

    def coerce_value(self, value):
        """
        Coerce the given value to the correct type/format for use with
        the filter.

        Default logic returns value as-is; subclass may override.
        """
        return value

    def filter_equal(self, query, value):
        """
        Filter data with an equal (``=``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query

        return query.filter(self.model_property == value)

    def filter_not_equal(self, query, value):
        """
        Filter data with a not equal (``!=``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query

        # sql probably excludes null values from results, but user
        # probably does not expect that, so explicitly include them.
        return query.filter(sa.or_(
            self.model_property == None,
            self.model_property != value,
        ))

    def filter_greater_than(self, query, value):
        """
        Filter data with a greater than (``>``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query
        return query.filter(self.model_property > value)

    def filter_greater_equal(self, query, value):
        """
        Filter data with a greater than or equal (``>=``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query
        return query.filter(self.model_property >= value)

    def filter_less_than(self, query, value):
        """
        Filter data with a less than (``<``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query
        return query.filter(self.model_property < value)

    def filter_less_equal(self, query, value):
        """
        Filter data with a less than or equal (``<=``) condition.
        """
        value = self.coerce_value(value)
        if value is None:
            return query
        return query.filter(self.model_property <= value)

    def filter_is_null(self, query, value):
        """
        Filter data with an ``IS NULL`` query.  The value is ignored.
        """
        return query.filter(self.model_property == None)

    def filter_is_not_null(self, query, value):
        """
        Filter data with an ``IS NOT NULL`` query.  The value is
        ignored.
        """
        return query.filter(self.model_property != None)


class StringAlchemyFilter(AlchemyFilter):
    """
    SQLAlchemy filter option for a text data column.

    Subclass of :class:`AlchemyFilter`.
    """
    default_verbs = ['contains', 'does_not_contain',
                     'equal', 'not_equal']

    def coerce_value(self, value):
        """ """
        if value is not None:
            value = str(value)
            if value:
                return value

    def filter_contains(self, query, value):
        """
        Filter data with an ``ILIKE`` condition.
        """
        value = self.coerce_value(value)
        if not value:
            return query

        criteria = []
        for val in value.split():
            val = val.replace('_', r'\_')
            val = f'%{val}%'
            criteria.append(self.model_property.ilike(val))

        return query.filter(sa.and_(*criteria))

    def filter_does_not_contain(self, query, value):
        """
        Filter data with a ``NOT ILIKE`` condition.
        """
        value = self.coerce_value(value)
        if not value:
            return query

        criteria = []
        for val in value.split():
            val = val.replace('_', r'\_')
            val = f'%{val}%'
            criteria.append(~self.model_property.ilike(val))

        # sql probably excludes null values from results, but user
        # probably does not expect that, so explicitly include them.
        return query.filter(sa.or_(
            self.model_property == None,
            sa.and_(*criteria)))


class NumericAlchemyFilter(AlchemyFilter):
    """
    SQLAlchemy filter option for a numeric data column.

    Subclass of :class:`AlchemyFilter`.
    """
    default_verbs = ['equal', 'not_equal',
                     'greater_than', 'greater_equal',
                     'less_than', 'less_equal']


class IntegerAlchemyFilter(NumericAlchemyFilter):
    """
    SQLAlchemy filter option for an integer data column.

    Subclass of :class:`NumericAlchemyFilter`.
    """

    def coerce_value(self, value):
        """ """
        if value:
            try:
                return int(value)
            except:
                pass


class BooleanAlchemyFilter(AlchemyFilter):
    """
    SQLAlchemy filter option for a boolean data column.

    Subclass of :class:`AlchemyFilter`.
    """
    default_verbs = ['is_true', 'is_false']

    def get_verbs(self):
        """ """

        # get basic verbs from caller, or default list
        verbs = getattr(self, 'verbs', self.default_verbs)
        if callable(verbs):
            verbs = verbs()
        verbs = list(verbs)

        # add some more if column is nullable
        if self.nullable:
            for verb in ('is_false_null', 'is_null', 'is_not_null'):
                if verb not in verbs:
                    verbs.append(verb)

        # add wildcard
        if 'is_any' not in verbs:
            verbs.append('is_any')

        return verbs

    def coerce_value(self, value):
        """ """
        if value is not None:
            return bool(value)

    def filter_is_true(self, query, value):
        """
        Filter data with an "is true" condition.  The value is
        ignored.
        """
        return query.filter(self.model_property == True)

    def filter_is_false(self, query, value):
        """
        Filter data with an "is false" condition.  The value is
        ignored.
        """
        return query.filter(self.model_property == False)

    def filter_is_false_null(self, query, value):
        """
        Filter data with "is false or null" condition.  The value is
        ignored.
        """
        return query.filter(sa.or_(self.model_property == False,
                                   self.model_property == None))


class DateAlchemyFilter(AlchemyFilter):
    """
    SQLAlchemy filter option for a
    :class:`sqlalchemy:sqlalchemy.types.Date` column.

    Subclass of :class:`AlchemyFilter`.
    """
    data_type = 'date'
    default_verbs = [
        'equal',
        'not_equal',
        'greater_than',
        'greater_equal',
        'less_than',
        'less_equal',
        # 'between',
    ]

    default_verb_labels = {
        'equal':                "on",
        'not_equal':            "not on",
        'greater_than':         "after",
        'greater_equal':        "on or after",
        'less_than':            "before",
        'less_equal':           "on or before",
        # 'between':              "between",
    }

    def coerce_value(self, value):
        """ """
        if value:
            if isinstance(value, datetime.date):
                return value

            try:
                dt = datetime.datetime.strptime(value, '%Y-%m-%d')
            except ValueError:
                log.warning("invalid date value: %s", value)
            else:
                return dt.date()


default_sqlalchemy_filters = {
    None: AlchemyFilter,
    sa.String: StringAlchemyFilter,
    sa.Text: StringAlchemyFilter,
    sa.Numeric: NumericAlchemyFilter,
    sa.Integer: IntegerAlchemyFilter,
    sa.Boolean: BooleanAlchemyFilter,
    sa.Date: DateAlchemyFilter,
}



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Progress Views
"""

from wuttaweb.progress import get_progress_session


def progress(request):
    """
    View which returns JSON with current progress status.

    The URL is like ``/progress/XXX`` where ``XXX`` is the "key" to a
    particular progress indicator, tied to a long-running operation.

    This key is used to lookup the progress status within the Beaker
    session storage.  See also
    :class:`~wuttaweb.progress.SessionProgress`.
    """
    key = request.matchdict['key']
    session = get_progress_session(request, key)

    # session has 'complete' flag set when operation is over
    if session.get('complete'):

        # set a flash msg for user if one is defined.  this is the
        # time to do it since user is about to get redirected.
        msg = session.get('success_msg')
        if msg:
            request.session.flash(msg)

    elif session.get('error'):  # uh-oh

        # set an error flash msg for user.  this is the time to do it
        # since user is about to get redirected.
        msg = session.get('error_msg', "An unspecified error occurred.")
        request.session.flash(msg, 'error')

    # nb. we return the session as-is; since it is dict-like (and only
    # contains relevant progress data) it can be used directly for the
    # JSON response context
    return session


def defaults(config, **kwargs):
    base = globals()

    progress = kwargs.get('progress', base['progress'])
    config.add_route('progress', '/progress/{key}')
    config.add_view(progress, route_name='progress', renderer='json')


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Report Views
"""

import datetime
import logging
import os
import tempfile

import deform

from wuttaweb.views import MasterView


log = logging.getLogger(__name__)


class ReportView(MasterView):
    """
    Master view for :term:`reports <report>`; route prefix is
    ``reports``.

    Notable URLs provided by this class:

    * ``/reports/``
    * ``/reports/XXX``
    """
    model_title = "Report"
    model_key = 'report_key'
    filterable = False
    sort_on_backend = False
    creatable = False
    editable = False
    deletable = False
    route_prefix = 'reports'
    template_prefix = '/reports'

    grid_columns = [
        'report_title',
        'help_text',
        'report_key',
    ]

    form_fields = [
        'help_text',
    ]

    def __init__(self, request, context=None):
        super().__init__(request, context=context)
        self.report_handler = self.app.get_report_handler()

    def get_grid_data(self, columns=None, session=None):
        """ """
        data = []
        for report in self.report_handler.get_reports().values():
            data.append(self.normalize_report(report))
        return data

    def normalize_report(self, report):
        """ """
        return {
            'report_key': report.report_key,
            'report_title': report.report_title,
            'help_text': report.__doc__,
        }

    def configure_grid(self, g):
        """ """
        super().configure_grid(g)

        # report_key
        g.set_link('report_key')

        # report_title
        g.set_link('report_title')
        g.set_searchable('report_title')

        # help_text
        g.set_searchable('help_text')

    def get_instance(self):
        """ """
        key = self.request.matchdict['report_key']
        report = self.report_handler.get_report(key)
        if report:
            return self.normalize_report(report)

        raise self.notfound()

    def get_instance_title(self, report):
        """ """
        return report['report_title']

    def view(self):
        """
        This lets user "view" the report but in this context that
        means showing them a form with report params, so they can run
        it.
        """
        key = self.request.matchdict['report_key']
        report = self.report_handler.get_report(key)
        normal = self.normalize_report(report)

        report_url = self.get_action_url('view', normal)
        form = self.make_model_form(normal,
                                    action_method='get',
                                    action_url=report_url,
                                    cancel_url=self.get_index_url(),
                                    show_button_reset=True,
                                    reset_url=report_url,
                                    button_label_submit="Run Report",
                                    button_icon_submit='arrow-circle-right')

        context = {
            'instance': normal,
            'report': report,
            'form': form,
            'xref_buttons': self.get_xref_buttons(report),
        }

        if self.request.GET:
            form.show_button_cancel = False
            context = self.run_report(report, context)

        return self.render_to_response('view', context)

    def configure_form(self, f):
        """ """
        super().configure_form(f)
        key = self.request.matchdict['report_key']
        report = self.report_handler.get_report(key)

        # help_text
        f.set_readonly('help_text')

        # add widget fields for all report params
        schema = f.get_schema()
        report.add_params(schema)
        f.set_fields([node.name for node in schema.children])

    def run_report(self, report, context):
        """
        Run the given report and update view template context.

        This is called automatically from :meth:`view()`.

        :param report:
           :class:`~wuttjamaican:wuttjamaican.reports.Report` instance
           to run.

        :param context: Current view template context.

        :returns: Final view template context.
        """
        form = context['form']
        controls = list(self.request.GET.items())

        # TODO: must re-inject help_text value for some reason,
        # otherwise its absence screws things up.  why?
        controls.append(('help_text', report.__doc__))

        dform = form.get_deform()
        try:
            params = dform.validate(controls)
        except deform.ValidationFailure:
            log.debug("form not valid: %s", dform.error)
            return context

        data = self.report_handler.make_report_data(report, params)

        columns = self.normalize_columns(report.get_output_columns())
        context['report_columns'] = columns

        format_cols = [col for col in columns if col.get('formatter')]
        if format_cols:
            for record in data['data']:
                for column in format_cols:
                    if column['name'] in record:
                        value = record[column['name']]
                        record[column['name']] = column['formatter'](value)

        params.pop('help_text')
        context['report_params'] = params
        context['report_data'] = data
        context['report_generated'] = datetime.datetime.now()
        return context

    def normalize_columns(self, columns):
        """ """
        normal = []
        for column in columns:
            if isinstance(column, str):
                column = {'name': column}
            column.setdefault('label', column['name'])
            normal.append(column)
        return normal

    def get_download_data(self):
        """ """
        key = self.request.matchdict['report_key']
        report = self.report_handler.get_report(key)
        params = dict(self.request.GET)
        columns = self.normalize_columns(report.get_output_columns())
        data = self.report_handler.make_report_data(report, params)
        return params, columns, data

    def get_download_path(self, data, ext):
        """ """
        tempdir = tempfile.mkdtemp()
        filename = f"{data['output_title']}.{ext}"
        return os.path.join(tempdir, filename)

    @classmethod
    def defaults(cls, config):
        """ """
        cls._defaults(config)
        cls._report_defaults(config)

    @classmethod
    def _report_defaults(cls, config):
        permission_prefix = cls.get_permission_prefix()
        model_title = cls.get_model_title()

        # overwrite title for "view" perm since it also implies "run"
        config.add_wutta_permission(permission_prefix,
                                    f'{permission_prefix}.view',
                                    f"View / run {model_title}")


        # separate permission to download report files
        config.add_wutta_permission(permission_prefix,
                                    f'{permission_prefix}.download',
                                    f"Download {model_title}")


def defaults(config, **kwargs):
    base = globals()

    ReportView = kwargs.get('ReportView', base['ReportView'])
    ReportView.defaults(config)


def includeme(config):
    defaults(config)
